"""Module docstring."""

import logging
import urllib.parse
from typing import Optional

from psycopg import errors as psycopg_errors

from ...dbstore import register_backend
from ...error import DBStoreError, DBStoreErrorCode
from ..errors import DatabaseError, DatabaseErrorCode
from .config import PostgresConfig
from .database import PostgresDatabase

LOGGER = logging.getLogger(__name__)


# Maps for compact error translation to reduce branching complexity
DB_ERROR_MAP = {
    DatabaseErrorCode.DATABASE_NOT_FOUND: (
        DBStoreErrorCode.NOT_FOUND,
        "Database Not Found",
    ),
    DatabaseErrorCode.PROFILE_NOT_FOUND: (
        DBStoreErrorCode.NOT_FOUND,
        "Database profile not found",
    ),
    DatabaseErrorCode.UNSUPPORTED_VERSION: (
        DBStoreErrorCode.UNSUPPORTED,
        "Unsupported release number in config table",
    ),
    DatabaseErrorCode.DEFAULT_PROFILE_NOT_FOUND: (
        DBStoreErrorCode.NOT_FOUND,
        "Database default profile not found",
    ),
    DatabaseErrorCode.CONNECTION_POOL_EXHAUSTED: (
        DBStoreErrorCode.UNEXPECTED,
        "Connection pool exhausted",
    ),
    DatabaseErrorCode.PROFILE_ALREADY_EXISTS: (
        DBStoreErrorCode.DUPLICATE,
        "Profile already exists",
    ),
    DatabaseErrorCode.RECORD_NOT_FOUND: (
        DBStoreErrorCode.NOT_FOUND,
        "Record not found",
    ),
    DatabaseErrorCode.DUPLICATE_ITEM_ENTRY_ERROR: (
        DBStoreErrorCode.DUPLICATE,
        "Duplicate Item Entry Error",
    ),
    DatabaseErrorCode.DATABASE_NOT_ENCRYPTED: (
        DBStoreErrorCode.UNEXPECTED,
        "Cannot rekey an unencrypted database",
    ),
    DatabaseErrorCode.CONNECTION_ERROR: (
        DBStoreErrorCode.UNEXPECTED,
        "Connection error",
    ),
    DatabaseErrorCode.QUERY_ERROR: (
        DBStoreErrorCode.UNEXPECTED,
        "Query error",
    ),
    DatabaseErrorCode.PROVISION_ERROR: (
        DBStoreErrorCode.UNEXPECTED,
        "Provision error",
    ),
}


class PostgresqlBackend:
    """PostgreSQL backend implementation for database manager."""

    async def provision(
        self,
        uri: str,
        _key_method: Optional[str],
        _pass_key: Optional[str],
        profile: Optional[str],
        recreate: bool,
        release_number: str,
        schema_config: Optional[str] = None,
        config: Optional[dict] = None,
    ):
        """Provision a new PostgreSQL database."""
        LOGGER.debug(
            "[provision_backend] uri=%s, profile=%s, recreate=%s, "
            "release_number=%s, schema_config=%s, config=%s",
            uri,
            profile,
            recreate,
            release_number,
            schema_config,
            config,
        )
        config = config or {}
        parsed_uri = urllib.parse.urlparse(uri)
        query_params = urllib.parse.parse_qs(parsed_uri.query)
        min_size = int(
            config.get("min_connections", query_params.get("min_connections", [4])[0])
        )
        max_size = int(
            config.get("max_connections", query_params.get("max_connections", [10])[0])
        )
        timeout = float(
            config.get("connect_timeout", query_params.get("connect_timeout", [30.0])[0])
        )
        max_idle = float(config.get("max_idle", query_params.get("max_idle", [5.0])[0]))
        max_lifetime = float(
            config.get("max_lifetime", query_params.get("max_lifetime", [3600.0])[0])
        )
        max_sessions = int(
            config.get(
                "max_sessions",
                query_params.get("max_sessions", [None])[0] or max_size * 0.75,
            )
        )
        config_obj = PostgresConfig(
            uri=uri,
            min_size=min_size,
            max_size=max_size,
            timeout=timeout,
            max_idle=max_idle,
            max_lifetime=max_lifetime,
            schema_config=schema_config or "generic",
        )
        (
            pool,
            profile_name,
            conn_str,
            effective_release_number,
        ) = await config_obj.provision(
            profile=profile, recreate=recreate, release_number=release_number
        )
        return PostgresDatabase(
            pool,
            profile_name,
            conn_str,
            effective_release_number,
            max_sessions=max_sessions,
            min_size=min_size,
            max_size=max_size,
            timeout=timeout,
            max_idle=max_idle,
            max_lifetime=max_lifetime,
            schema_context=config_obj.schema_context,
            backend=self,
        )

    async def open(
        self,
        uri: str,
        _key_method: Optional[str],
        _pass_key: Optional[str],
        profile: Optional[str],
        schema_migration: Optional[bool] = None,
        target_schema_release_number: Optional[str] = None,
        config: Optional[dict] = None,
    ):
        """Open an existing PostgreSQL database."""
        LOGGER.debug(
            "[open_backend] uri=%s, profile=%s, schema_migration=%s, "
            "target_schema_release_number=%s, config=%s, "
            "schema_config will be retrieved from database",
            uri,
            profile,
            schema_migration,
            target_schema_release_number,
            config,
        )
        config = config or {}
        parsed_uri = urllib.parse.urlparse(uri)
        query_params = urllib.parse.parse_qs(parsed_uri.query)
        min_size = int(
            config.get("min_connections", query_params.get("min_connections", [4])[0])
        )
        max_size = int(
            config.get("max_connections", query_params.get("max_connections", [10])[0])
        )
        timeout = float(
            config.get("connect_timeout", query_params.get("connect_timeout", [30.0])[0])
        )
        max_idle = float(config.get("max_idle", query_params.get("max_idle", [5.0])[0]))
        max_lifetime = float(
            config.get("max_lifetime", query_params.get("max_lifetime", [3600.0])[0])
        )
        max_sessions = int(
            config.get(
                "max_sessions",
                query_params.get("max_sessions", [None])[0] or max_size * 0.75,
            )
        )
        config_obj = PostgresConfig(
            uri=uri,
            min_size=min_size,
            max_size=max_size,
            timeout=timeout,
            max_idle=max_idle,
            max_lifetime=max_lifetime,
        )
        pool, profile_name, conn_str, effective_release_number = await config_obj.open(
            profile=profile,
            schema_migration=schema_migration,
            target_schema_release_number=target_schema_release_number,
        )
        return PostgresDatabase(
            pool,
            profile_name,
            conn_str,
            effective_release_number,
            max_sessions=max_sessions,
            min_size=min_size,
            max_size=max_size,
            timeout=timeout,
            max_idle=max_idle,
            max_lifetime=max_lifetime,
            schema_context=config_obj.schema_context,
            backend=self,
        )

    async def remove(self, uri: str, config: Optional[dict] = None):
        """Remove a PostgreSQL database."""
        LOGGER.debug("[remove_backend] Starting with uri=%s, config=%s", uri, config)
        config = config or {}
        parsed_uri = urllib.parse.urlparse(uri)
        query_params = urllib.parse.parse_qs(parsed_uri.query)
        min_size = int(
            config.get("min_connections", query_params.get("min_connections", [4])[0])
        )
        max_size = int(
            config.get("max_connections", query_params.get("max_connections", [10])[0])
        )
        timeout = float(
            config.get("connect_timeout", query_params.get("connect_timeout", [30.0])[0])
        )
        max_idle = float(config.get("max_idle", query_params.get("max_idle", [5.0])[0]))
        max_lifetime = float(
            config.get("max_lifetime", query_params.get("max_lifetime", [3600.0])[0])
        )
        config_obj = PostgresConfig(
            uri=uri,
            min_size=min_size,
            max_size=max_size,
            timeout=timeout,
            max_idle=max_idle,
            max_lifetime=max_lifetime,
            schema_config="generic",
        )
        result = await config_obj.remove()
        return result

    def translate_error(self, error: Exception) -> DBStoreError:
        """Translate database errors to DBStoreError."""
        LOGGER.debug("Translating error: %s, type=%s", str(error), type(error))
        if isinstance(error, DatabaseError):
            mapped = DB_ERROR_MAP.get(error.code)
            if mapped:
                return DBStoreError(code=mapped[0], message=mapped[1])
        elif isinstance(error, psycopg_errors.UniqueViolation):
            return DBStoreError(
                code=DBStoreErrorCode.DUPLICATE, message=f"Duplicate entry: {str(error)}"
            )
        elif isinstance(error, psycopg_errors.ForeignKeyViolation):
            return DBStoreError(
                code=DBStoreErrorCode.UNEXPECTED,
                message=f"Foreign key violation: {str(error)}",
            )
        elif isinstance(error, psycopg_errors.OperationalError):
            return DBStoreError(
                code=DBStoreErrorCode.BACKEND,
                message=f"Database operation failed: {str(error)}",
            )
        elif isinstance(error, TypeError):
            return DBStoreError(
                code=DBStoreErrorCode.UNEXPECTED,
                message=f"Configuration error: {str(error)}",
            )
        return DBStoreError(
            code=DBStoreErrorCode.UNEXPECTED, message=f"Unexpected error: {str(error)}"
        )


register_backend("postgresql", PostgresqlBackend())



"""AnonCreds revocation list routes."""

import logging
from asyncio import shield

from aiohttp import web
from aiohttp_apispec import docs, request_schema, response_schema

from .....admin.decorators.auth import tenant_authentication
from ....models.revocation import RevListResultSchema
from ....revocation.revocation import AnonCredsRevocation
from ...common.utils import get_request_body_with_profile_check
from .. import REVOCATION_TAG_TITLE
from .models import RevListCreateRequestSchema

LOGGER = logging.getLogger(__name__)


@docs(
    tags=[REVOCATION_TAG_TITLE],
    summary="Create and publish a revocation status list on the connected datastore",
)
@request_schema(RevListCreateRequestSchema())
@response_schema(RevListResultSchema(), 200, description="")
@tenant_authentication
async def rev_list_post(request: web.BaseRequest):
    """Request handler for creating registering a revocation list."""
    _, profile, body, options = await get_request_body_with_profile_check(request)

    rev_reg_def_id = body["rev_reg_def_id"]  # required in request schema

    revocation = AnonCredsRevocation(profile)
    result = await shield(
        revocation.create_and_register_revocation_list(rev_reg_def_id, options=options)
    )
    if isinstance(result, str):  # if it's a string, it's an error message
        raise web.HTTPBadRequest(reason=result)

    LOGGER.debug("published revocation list for: %s", rev_reg_def_id)
    return web.json_response(result.serialize())


async def register(app: web.Application) -> None:
    """Register routes."""
    app.add_routes(
        [
            web.post("/anoncreds/revocation-list", rev_list_post),
        ]
    )



# """Test cases for MongoTagEncoder with In and Exist conjunctions."""

import json
import unittest

from acapy_agent.database_manager.wql_nosql.encoders import encoder_factory
from acapy_agent.database_manager.wql_nosql.tags import TagName, TagQuery


class TestMongoTagEncoder(unittest.TestCase):
    """Test cases for the MongoTagEncoder class."""

    def test_in_and_exist_conjunction(self):
        """Test encoding an In and Exist conjunction into MongoDB query."""
        # Define the query: color in ['red', 'blue'] AND size exists
        query = TagQuery.and_(
            [
                TagQuery.in_(TagName("color"), ["red", "blue"]),
                TagQuery.exist([TagName("size")]),
            ]
        )

        # Set up encoding functions
        def enc_name(x):
            return x  # No transformation for names

        def enc_value(x):
            return x  # No transformation for values

        # Get the encoder for MongoDB
        encoder = encoder_factory.get_encoder("mongodb", enc_name, enc_value)

        # Encode the query
        mongo_query = encoder.encode_query(query)

        # Print the generated query for debugging
        print("\nGenerated MongoDB Query:")
        print(json.dumps(mongo_query, indent=2))

        # Define the expected MongoDB query
        expected_query = {
            "$and": [{"color": {"$in": ["red", "blue"]}}, {"size": {"$exists": True}}]
        }

        # Print the expected query for comparison
        print("\nExpected MongoDB Query:")
        print(json.dumps(expected_query, indent=2))

        # Assert that the generated query matches the expected query
        self.assertEqual(mongo_query, expected_query)

        # Instructions for manual testing with mongosh
        print("\n### Manual Testing Instructions with mongosh")
        print("To verify the query manually, follow these steps:")
        print("1. Open a terminal and start mongosh:")
        print("   ```bash")
        print("   mongosh")
        print("   ```")
        print("2. Switch to or create a test database:")
        print("   ```javascript")
        print("   use test_db")
        print("   ```")
        print("3. Create a collection and insert sample documents:")
        print("   ```javascript")
        print("   db.items.drop()  // Clear existing collection")
        print("   db.items.insertMany([")
        print("       { _id: 1, color: 'red', size: 'M' },")
        print("       { _id: 2, color: 'blue' },")
        print("       { _id: 3, color: 'green', size: 'L' },")
        print("       { _id: 4, size: 'S' },")
        print("       { _id: 5, color: 'blue', size: 'S' }")
        print("   ])")
        print("   ```")
        print("4. Run the generated query:")
        print("   ```javascript")
        print(f"   db.items.find({json.dumps(mongo_query)})")
        print("   ```")
        print("5. Expected result: Documents with _id: 1 and 5")
        print("   - _id: 1: color='red' (in ['red', 'blue']), size='M' (exists)")
        print("   - _id: 5: color='blue' (in ['red', 'blue']), size='S' (exists)")
        print("6. Clean up (optional):")
        print("   ```javascript")
        print("   db.items.drop()")
        print("   ```")


if __name__ == "__main__":
    unittest.main()



"""Module docstring."""

import asyncio
import logging
import urllib.parse

from psycopg_pool import AsyncConnectionPool

from ..errors import DatabaseError, DatabaseErrorCode

LOGGER = logging.getLogger(__name__)


class PostgresConnectionPool:
    """Connection pool manager for PostgreSQL databases."""

    def __init__(
        self,
        conn_str: str,
        min_size: int = 4,
        max_size: int = 100,
        timeout: float = 30.0,
        max_idle: float = 5.0,
        max_lifetime: float = 3600.0,
    ):
        """Initialize PostgreSQL connection pool."""
        # Sanitize connection string by removing admin parameters
        self.conn_str = self._sanitize_conn_str(conn_str)
        self.min_size = min_size
        self.max_size = max_size
        self.timeout = timeout
        self.max_idle = max_idle
        self.max_lifetime = max_lifetime
        self.pool = None
        self.connection_count = 0
        self.connection_ids = {}

    def _sanitize_conn_str(self, conn_str: str) -> str:
        """Remove admin parameters from connection string that psycopg doesn't accept."""
        parsed = urllib.parse.urlparse(conn_str)
        query_params = urllib.parse.parse_qs(parsed.query)

        # Remove admin parameters
        admin_params = ["admin_account", "admin_password"]
        for param in admin_params:
            query_params.pop(param, None)

        # Rebuild query string
        new_query = urllib.parse.urlencode(query_params, doseq=True)

        # Rebuild URL
        sanitized = parsed._replace(query=new_query).geturl()
        return sanitized

    async def initialize(self):
        """Initialize the connection pool."""
        try:
            self.pool = AsyncConnectionPool(
                conninfo=self.conn_str,
                min_size=self.min_size,
                max_size=self.max_size,
                timeout=self.timeout,
                max_idle=self.max_idle,
                max_lifetime=self.max_lifetime,
                kwargs={"options": "-c client_encoding=UTF8"},  # Ensure UTF-8 encoding
            )
            await self.pool.open()  # Explicitly open pool to avoid deprecation warning
            await self.pool.wait()
            LOGGER.debug("Connection pool initialized with %d connections", self.min_size)
        except Exception as e:
            LOGGER.error("Failed to initialize connection pool: %s", str(e))
            raise DatabaseError(
                code=DatabaseErrorCode.CONNECTION_ERROR,
                message="Failed to initialize connection pool",
                actual_error=str(e),
            )

    async def getconn(self):
        """Get a connection from the pool."""
        try:
            conn = await asyncio.wait_for(self.pool.getconn(), timeout=60.0)
            # Rollback any existing transaction to ensure clean state
            # This ensures the connection is in IDLE state before returning
            await conn.rollback()
            # Note: UTF-8 encoding is already set at pool creation via kwargs
            # (options: -c client_encoding=UTF8), so no need to SET it here.
            # Executing SET here would start an implicit transaction and add latency.
            conn_id = self.connection_count
            self.connection_ids[id(conn)] = conn_id
            self.connection_count += 1
            LOGGER.debug(
                "Connection ID=%d retrieved from pool. Pool size: %d/%d",
                conn_id,
                self.pool.get_stats().get("pool_available", 0),
                self.max_size,
            )
            return conn
        except asyncio.TimeoutError as e:
            LOGGER.error("Failed to retrieve connection from pool: %s", str(e))
            raise DatabaseError(
                code=DatabaseErrorCode.CONNECTION_POOL_EXHAUSTED,
                message="Connection pool exhausted after 60.0 seconds",
                actual_error=str(e),
            )
        except Exception as e:
            LOGGER.error("Failed to retrieve connection from pool: %s", str(e))
            raise DatabaseError(
                code=DatabaseErrorCode.CONNECTION_POOL_EXHAUSTED,
                message="Connection pool exhausted",
                actual_error=str(e),
            )

    async def putconn(self, conn):
        """Return a connection to the pool."""
        try:
            # Roll back any open transactions to ensure clean state
            await conn.rollback()
            await self.pool.putconn(conn)
            conn_id = self.connection_ids.pop(id(conn), -1)
            LOGGER.debug(
                "Connection ID=%d returned to pool. Pool size: %d/%d",
                conn_id,
                self.pool.get_stats().get("pool_available", 0),
                self.max_size,
            )
        except Exception as e:
            LOGGER.error("Failed to return connection to pool: %s", str(e))
            raise DatabaseError(
                code=DatabaseErrorCode.CONNECTION_ERROR,
                message="Failed to return connection to pool",
                actual_error=str(e),
            )

    async def close(self):
        """Close the connection pool."""
        try:
            await self.pool.close()
            self.connection_ids.clear()
            LOGGER.debug("Connection pool closed")
        except Exception as e:
            LOGGER.error("Failed to close connection pool: %s", str(e))
            raise DatabaseError(
                code=DatabaseErrorCode.CONNECTION_ERROR,
                message="Failed to close connection pool",
                actual_error=str(e),
            )



"""Module docstring."""

import logging
import sqlite3
from typing import Optional

from ...dbstore import DatabaseBackend
from ...error import DBStoreError, DBStoreErrorCode
from ..errors import DatabaseError, DatabaseErrorCode
from .config import SqliteConfig
from .database import SqliteDatabase

LOGGER = logging.getLogger(__name__)


class SqliteBackend(DatabaseBackend):
    """SQLite backend implementation for database manager."""

    def provision(
        self,
        uri: str,
        key_method: Optional[str],
        pass_key: Optional[str],
        profile: Optional[str],
        recreate: bool,
        release_number: str = "release_0",
        schema_config: str = "generic",
        config: Optional[dict] = None,
    ):
        """Provision a new SQLite database instance.

        Uses specified release number and schema config.
        """
        LOGGER.debug(
            "[provision_backend] Starting with uri=%s, profile=%s, recreate=%s, "
            "release_number=%s, schema_config=%s, config=%s",
            uri,
            profile,
            recreate,
            release_number,
            schema_config,
            config,
        )
        config_obj = SqliteConfig(
            uri=uri, encryption_key=pass_key, schema_config=schema_config
        )
        pool, profile_name, path, effective_release_number = config_obj.provision(
            profile=profile, recreate=recreate, release_number=release_number
        )
        return SqliteDatabase(pool, profile_name, path, effective_release_number)

    def open(
        self,
        uri: str,
        key_method: Optional[str],
        pass_key: Optional[str],
        profile: Optional[str],
        schema_migration: Optional[bool] = None,
        target_schema_release_number: Optional[str] = None,
        schema_config: Optional[str] = None,
        config: Optional[dict] = None,
    ):
        """Open an existing SQLite database instance with optional migration."""
        LOGGER.debug(
            "[open_backend] Starting with uri=%s, profile=%s, schema_migration=%s, "
            "target_schema_release_number=%s, config=%s",
            uri,
            profile,
            schema_migration,
            target_schema_release_number,
            config,
        )
        config_obj = SqliteConfig(uri=uri, encryption_key=pass_key)
        pool, profile_name, path, effective_release_number = config_obj.open(
            profile=profile,
            schema_migration=schema_migration,
            target_schema_release_number=target_schema_release_number,
        )
        return SqliteDatabase(pool, profile_name, path, effective_release_number)

    def remove(
        self, uri: str, release_number: str = "release_0", config: Optional[dict] = None
    ):
        """Remove the SQLite database file."""
        LOGGER.debug(
            "[remove_backend] Starting with uri=%s, release_number=%s, config=%s",
            uri,
            release_number,
            config,
        )
        config_obj = SqliteConfig(uri=uri)
        result = config_obj.remove()
        return result

    def translate_error(self, exception):
        """Translate backend-specific exceptions to DBStoreError."""
        # Map DatabaseError codes to DBStoreError
        database_error_mapping = {
            DatabaseErrorCode.DATABASE_NOT_FOUND: (
                DBStoreErrorCode.NOT_FOUND,
                "Database Not Found",
            ),
            DatabaseErrorCode.UNSUPPORTED_VERSION: (
                DBStoreErrorCode.UNSUPPORTED,
                "Unsupported release number in config table",
            ),
            DatabaseErrorCode.DEFAULT_PROFILE_NOT_FOUND: (
                DBStoreErrorCode.NOT_FOUND,
                "Database default profile not found",
            ),
            DatabaseErrorCode.PROFILE_NOT_FOUND: (
                DBStoreErrorCode.NOT_FOUND,
                "Database profile not found",
            ),
            DatabaseErrorCode.CONNECTION_POOL_EXHAUSTED: (
                DBStoreErrorCode.UNEXPECTED,
                "Connection pool exhausted",
            ),
            DatabaseErrorCode.PROFILE_ALREADY_EXISTS: (
                DBStoreErrorCode.DUPLICATE,
                "Profile already exists",
            ),
            DatabaseErrorCode.RECORD_NOT_FOUND: (
                DBStoreErrorCode.NOT_FOUND,
                "Record not found",
            ),
            DatabaseErrorCode.DUPLICATE_ITEM_ENTRY_ERROR: (
                DBStoreErrorCode.DUPLICATE,
                "Duplicate Item Entry Error",
            ),
            DatabaseErrorCode.DATABASE_NOT_ENCRYPTED: (
                DBStoreErrorCode.UNEXPECTED,
                "Cannot rekey an unencrypted database",
            ),
            DatabaseErrorCode.CONNECTION_ERROR: (
                DBStoreErrorCode.UNEXPECTED,
                "Connection error",
            ),
            DatabaseErrorCode.QUERY_ERROR: (DBStoreErrorCode.UNEXPECTED, "Query error"),
            DatabaseErrorCode.PROVISION_ERROR: (
                DBStoreErrorCode.UNEXPECTED,
                "Provision error",
            ),
        }

        if isinstance(exception, DatabaseError):
            mapping = database_error_mapping.get(exception.code)
            if mapping:
                return DBStoreError(code=mapping[0], message=mapping[1])
        elif isinstance(exception, sqlite3.IntegrityError):
            return DBStoreError(
                code=DBStoreErrorCode.DUPLICATE, message="Duplicate entry"
            )
        elif isinstance(exception, sqlite3.OperationalError):
            return DBStoreError(
                code=DBStoreErrorCode.BACKEND, message="Database operation failed"
            )

        return DBStoreError(
            code=DBStoreErrorCode.UNEXPECTED,
            message="Unexpected error",
            extra=str(exception),
        )



"""Module docstring."""

CATEGORY = "issuer_cred_rev"

IDX_ISSUER_CRED_REV_ON_ITEM_ID = "ON issuer_cred_rev_v0_1 (item_id);"
IDX_ISSUER_CRED_REV_ON_CRED_EX_ID = "ON issuer_cred_rev_v0_1 (cred_ex_id);"
IDX_ISSUER_CRED_REV_ON_REV_REG_ID = "ON issuer_cred_rev_v0_1 (rev_reg_id);"
IDX_ISSUER_CRED_REV_ON_CRED_DEF_ID = "ON issuer_cred_rev_v0_1 (cred_def_id);"
IDX_ISSUER_CRED_REV_ON_STATE = "ON issuer_cred_rev_v0_1 (state);"
IDX_ISSUER_CRED_REV_ON_REV_REG_CRED_REV = (
    "ON issuer_cred_rev_v0_1 (rev_reg_id, cred_rev_id);"
)

SCHEMAS = {
    "sqlite": [
        """
        CREATE TABLE IF NOT EXISTS issuer_cred_rev_v0_1 (
            id INTEGER PRIMARY KEY AUTOINCREMENT,
            item_id INTEGER NOT NULL,
            item_name TEXT NOT NULL,
            state TEXT,
            cred_ex_id TEXT,
            rev_reg_id TEXT,
            cred_rev_id TEXT,
            cred_def_id TEXT,
            cred_ex_version TEXT,
            created_at DATETIME DEFAULT CURRENT_TIMESTAMP,
            updated_at DATETIME DEFAULT CURRENT_TIMESTAMP,
            FOREIGN KEY (item_id) REFERENCES items(id) 
                ON DELETE CASCADE ON UPDATE CASCADE,
            CONSTRAINT issuer_cred_rev_v0_1_unique_rev_reg_cred_rev 
                UNIQUE (rev_reg_id, cred_rev_id)
        );
        """,
        "CREATE INDEX IF NOT EXISTS idx_issuer_cred_rev_item_id_v0_1 "
        + IDX_ISSUER_CRED_REV_ON_ITEM_ID,
        "CREATE INDEX IF NOT EXISTS idx_issuer_cred_rev_cred_ex_id_v0_1 "
        + IDX_ISSUER_CRED_REV_ON_CRED_EX_ID,
        "CREATE INDEX IF NOT EXISTS idx_issuer_cred_rev_rev_reg_id_v0_1 "
        + IDX_ISSUER_CRED_REV_ON_REV_REG_ID,
        "CREATE INDEX IF NOT EXISTS idx_issuer_cred_rev_cred_def_id_v0_1 "
        + IDX_ISSUER_CRED_REV_ON_CRED_DEF_ID,
        "CREATE INDEX IF NOT EXISTS idx_issuer_cred_rev_state_v0_1 "
        + IDX_ISSUER_CRED_REV_ON_STATE,
        "CREATE INDEX IF NOT EXISTS idx_issuer_cred_rev_rev_reg_cred_rev_v0_1 "
        + IDX_ISSUER_CRED_REV_ON_REV_REG_CRED_REV,
        """
        CREATE TRIGGER IF NOT EXISTS trg_update_issuer_cred_rev_timestamp_v0_1
        AFTER UPDATE ON issuer_cred_rev_v0_1
        FOR EACH ROW
        BEGIN
            UPDATE issuer_cred_rev_v0_1
            SET updated_at = CURRENT_TIMESTAMP
            WHERE id = OLD.id;
        END;
        """,
    ],
    "postgresql": [
        """
        CREATE TABLE IF NOT EXISTS issuer_cred_rev_v0_1 (
            id SERIAL PRIMARY KEY,
            item_id INTEGER NOT NULL,
            item_name TEXT NOT NULL,
            state TEXT,
            cred_ex_id TEXT,
            rev_reg_id TEXT,
            cred_rev_id TEXT,
            cred_def_id TEXT,
            cred_ex_version TEXT,
            created_at TIMESTAMP WITH TIME ZONE DEFAULT CURRENT_TIMESTAMP,
            updated_at TIMESTAMP WITH TIME ZONE DEFAULT CURRENT_TIMESTAMP,
            CONSTRAINT fk_item_id FOREIGN KEY (item_id) REFERENCES items(id) 
                ON DELETE CASCADE ON UPDATE CASCADE,
            CONSTRAINT issuer_cred_rev_v0_1_unique_rev_reg_cred_rev 
                UNIQUE (rev_reg_id, cred_rev_id)
        );
        """,
        "CREATE INDEX IF NOT EXISTS idx_issuer_cred_rev_item_id_v0_1 "
        + IDX_ISSUER_CRED_REV_ON_ITEM_ID,
        "CREATE INDEX IF NOT EXISTS idx_issuer_cred_rev_cred_ex_id_v0_1 "
        + IDX_ISSUER_CRED_REV_ON_CRED_EX_ID,
        "CREATE INDEX IF NOT EXISTS idx_issuer_cred_rev_rev_reg_id_v0_1 "
        + IDX_ISSUER_CRED_REV_ON_REV_REG_ID,
        "CREATE INDEX IF NOT EXISTS idx_issuer_cred_rev_cred_def_id_v0_1 "
        + IDX_ISSUER_CRED_REV_ON_CRED_DEF_ID,
        "CREATE INDEX IF NOT EXISTS idx_issuer_cred_rev_state_v0_1 "
        + IDX_ISSUER_CRED_REV_ON_STATE,
        "CREATE INDEX IF NOT EXISTS idx_issuer_cred_rev_rev_reg_cred_rev_v0_1 "
        + IDX_ISSUER_CRED_REV_ON_REV_REG_CRED_REV,
        """
        CREATE OR REPLACE FUNCTION update_issuer_cred_rev_timestamp_v0_1()
        RETURNS TRIGGER AS $$
        BEGIN
            IF NEW.updated_at IS NULL THEN
                NEW.updated_at = CURRENT_TIMESTAMP;
            END IF;
            RETURN NEW;
        END;
        $$ LANGUAGE plpgsql;
        """,
        """
        CREATE TRIGGER trg_update_issuer_cred_rev_timestamp_v0_1
        BEFORE UPDATE ON issuer_cred_rev_v0_1
        FOR EACH ROW
        EXECUTE FUNCTION update_issuer_cred_rev_timestamp_v0_1();
        """,
    ],
    "mssql": [
        """
        CREATE TABLE issuer_cred_rev_v0_1 (
            id INT IDENTITY(1,1) PRIMARY KEY,
            item_id INT NOT NULL,
            item_name NVARCHAR(MAX) NOT NULL,
            state NVARCHAR(255),
            cred_ex_id NVARCHAR(255),
            rev_reg_id NVARCHAR(255),
            cred_rev_id NVARCHAR(255),
            cred_def_id NVARCHAR(255),
            cred_ex_version NVARCHAR(50),
            created_at DATETIME2 DEFAULT SYSDATETIME(),
            updated_at DATETIME2 DEFAULT SYSDATETIME(),
            CONSTRAINT fk_item_id FOREIGN KEY (item_id) REFERENCES items(id) 
                ON DELETE CASCADE ON UPDATE CASCADE,
            CONSTRAINT issuer_cred_rev_v0_1_unique_rev_reg_cred_rev 
                UNIQUE (rev_reg_id, cred_rev_id)
        );
        """,
        "CREATE NONCLUSTERED INDEX idx_issuer_cred_rev_item_id_v0_1 "
        + IDX_ISSUER_CRED_REV_ON_ITEM_ID,
        "CREATE NONCLUSTERED INDEX idx_issuer_cred_rev_cred_ex_id_v0_1 "
        + IDX_ISSUER_CRED_REV_ON_CRED_EX_ID,
        "CREATE NONCLUSTERED INDEX idx_issuer_cred_rev_rev_reg_id_v0_1 "
        + IDX_ISSUER_CRED_REV_ON_REV_REG_ID,
        "CREATE NONCLUSTERED INDEX idx_issuer_cred_rev_cred_def_id_v0_1 "
        + IDX_ISSUER_CRED_REV_ON_CRED_DEF_ID,
        "CREATE NONCLUSTERED INDEX idx_issuer_cred_rev_state_v0_1 "
        + IDX_ISSUER_CRED_REV_ON_STATE,
        "CREATE NONCLUSTERED INDEX idx_issuer_cred_rev_rev_reg_cred_rev_v0_1 "
        + IDX_ISSUER_CRED_REV_ON_REV_REG_CRED_REV,
        """
        CREATE TRIGGER trg_update_issuer_cred_rev_timestamp_v0_1
        ON issuer_cred_rev_v0_1
        AFTER UPDATE
        AS
        BEGIN
            UPDATE issuer_cred_rev_v0_1
            SET updated_at = SYSDATETIME()
            FROM issuer_cred_rev_v0_1
            INNER JOIN inserted ON issuer_cred_rev_v0_1.id = inserted.id
            WHERE inserted.updated_at IS NULL;
        END;
        """,
    ],
}

DROP_SCHEMAS = {
    "sqlite": [
        "DROP TRIGGER IF EXISTS trg_update_issuer_cred_rev_timestamp_v0_1;",
        "DROP INDEX IF EXISTS idx_issuer_cred_rev_rev_reg_cred_rev_v0_1;",
        "DROP INDEX IF EXISTS idx_issuer_cred_rev_state_v0_1;",
        "DROP INDEX IF EXISTS idx_issuer_cred_rev_cred_def_id_v0_1;",
        "DROP INDEX IF EXISTS idx_issuer_cred_rev_rev_reg_id_v0_1;",
        "DROP INDEX IF EXISTS idx_issuer_cred_rev_cred_ex_id_v0_1;",
        "DROP INDEX IF EXISTS idx_issuer_cred_rev_item_id_v0_1;",
        "DROP TABLE IF EXISTS issuer_cred_rev_v0_1;",
    ],
    "postgresql": [
        "DROP TRIGGER IF EXISTS trg_update_issuer_cred_rev_timestamp_v0_1 "
        "ON issuer_cred_rev_v0_1;",
        "DROP FUNCTION IF EXISTS update_issuer_cred_rev_timestamp_v0_1 CASCADE;",
        "DROP INDEX IF EXISTS idx_issuer_cred_rev_rev_reg_cred_rev_v0_1;",
        "DROP INDEX IF EXISTS idx_issuer_cred_rev_state_v0_1;",
        "DROP INDEX IF EXISTS idx_issuer_cred_rev_cred_def_id_v0_1;",
        "DROP INDEX IF EXISTS idx_issuer_cred_rev_rev_reg_id_v0_1;",
        "DROP INDEX IF EXISTS idx_issuer_cred_rev_cred_ex_id_v0_1;",
        "DROP INDEX IF EXISTS idx_issuer_cred_rev_item_id_v0_1;",
        "DROP TABLE IF EXISTS issuer_cred_rev_v0_1 CASCADE;",
    ],
    "mssql": [
        "DROP TRIGGER IF EXISTS trg_update_issuer_cred_rev_timestamp_v0_1;",
        "DROP INDEX IF EXISTS idx_issuer_cred_rev_rev_reg_cred_rev_v0_1 "
        "ON issuer_cred_rev_v0_1;",
        "DROP INDEX IF EXISTS idx_issuer_cred_rev_state_v0_1 ON issuer_cred_rev_v0_1;",
        "DROP INDEX IF EXISTS idx_issuer_cred_rev_cred_def_id_v0_1 "
        "ON issuer_cred_rev_v0_1;",
        "DROP INDEX IF EXISTS idx_issuer_cred_rev_rev_reg_id_v0_1 "
        "ON issuer_cred_rev_v0_1;",
        "DROP INDEX IF EXISTS idx_issuer_cred_rev_cred_ex_id_v0_1 "
        "ON issuer_cred_rev_v0_1;",
        "DROP INDEX IF EXISTS idx_issuer_cred_rev_item_id_v0_1 ON issuer_cred_rev_v0_1;",
        "DROP TABLE IF EXISTS issuer_cred_rev_v0_1;",
    ],
}


COLUMNS = [
    "state",
    "cred_ex_id",
    "rev_reg_id",
    "cred_rev_id",
    "cred_def_id",
    "cred_ex_version",
    "created_at",
    "updated_at",
]

# sample
# category=issuer_cred_rev, name=76db16bc-bcfb-4d91-8c89-53373f09bd4a,
# Sample issuer credential revocation record:
# value={
#   "cred_ex_id": "e8a39578-b7e3-4682-b319-d2f5433adf25",
#   "cred_rev_id": "1",
#   "cred_ex_version": "2",
#   "cred_def_id": "BacujJ3zNmAR9afs9hPryb:3:CL:2842581:cd0.31",
#   "rev_reg_id": "BacujJ3zNmAR9afs9hPryb:4:...:CL_ACCUM:0",
#   "state": "issued",
#   "created_at": "2025-06-17T19:29:48.947936Z",
#   "updated_at": "2025-06-17T19:29:48.947936Z"
# },
# tags={
#   'cred_ex_id': 'e8a39578-b7e3-4682-b319-d2f5433adf25',
#   'cred_rev_id': '1',
#   'cred_ex_version': '2',
#   'cred_def_id': 'BacujJ3zNmAR9afs9hPryb:3:CL:2842581:cd0.31',
#   'rev_reg_id': 'BacujJ3zNmAR9afs9hPryb:4:...:CL_ACCUM:0',
#   'state': 'issued'
# }, expiry_ms=None, value_json=None



"""Module docstring."""

CATEGORY = "connection"

SCHEMAS = {
    "sqlite": [
        """
        CREATE TABLE IF NOT EXISTS connection_v0_1 (
            id INTEGER PRIMARY KEY AUTOINCREMENT,
            item_id INTEGER NOT NULL,
            item_name TEXT,
            state TEXT,
            created_at DATETIME DEFAULT CURRENT_TIMESTAMP,
            updated_at DATETIME DEFAULT CURRENT_TIMESTAMP,
            my_did TEXT,
            their_did TEXT,
            their_label TEXT,
            their_role TEXT,
            invitation_key TEXT,
            invitation_msg_id TEXT,
            request_id TEXT,
            inbound_connection_id TEXT,
            error_msg TEXT,
            accept TEXT,
            invitation_mode TEXT,
            alias TEXT,
            their_public_did TEXT,
            connection_protocol TEXT,
            rfc23_state TEXT,
            FOREIGN KEY (item_id) REFERENCES items(id) ON DELETE CASCADE ON UPDATE CASCADE
        );
        """,
        """
        CREATE INDEX IF NOT EXISTS idx_connection_item_id_v0_1
        ON connection_v0_1 (item_id);
        """,
        """
        CREATE INDEX IF NOT EXISTS idx_connection_state_v0_1
        ON connection_v0_1 (state);
        """,
        """
        CREATE INDEX IF NOT EXISTS idx_connection_created_at_v0_1
        ON connection_v0_1 (created_at);
        """,
        """
        CREATE TRIGGER IF NOT EXISTS trg_update_connection_timestamp_v0_1
        AFTER UPDATE ON connection_v0_1
        FOR EACH ROW
        BEGIN
            UPDATE connection_v0_1
            SET updated_at = CURRENT_TIMESTAMP
            WHERE id = OLD.id;
        END;
        """,
    ],
    "postgresql": [
        """
        CREATE TABLE IF NOT EXISTS connection_v0_1 (
            id SERIAL PRIMARY KEY,
            item_id INTEGER NOT NULL,
            item_name TEXT,
            state TEXT,
            created_at TIMESTAMP WITH TIME ZONE DEFAULT CURRENT_TIMESTAMP,
            updated_at TIMESTAMP WITH TIME ZONE DEFAULT CURRENT_TIMESTAMP,
            my_did TEXT,
            their_did TEXT,
            their_label TEXT,
            their_role TEXT,
            invitation_key TEXT,
            invitation_msg_id TEXT,
            request_id TEXT,
            inbound_connection_id TEXT,
            error_msg TEXT,
            accept TEXT,
            invitation_mode TEXT,
            alias TEXT,
            their_public_did TEXT,
            connection_protocol TEXT,
            rfc23_state TEXT,
            CONSTRAINT fk_item_id FOREIGN KEY (item_id)
            REFERENCES items(id) ON DELETE CASCADE ON UPDATE CASCADE
        );
        """,
        """
        CREATE INDEX IF NOT EXISTS idx_connection_item_id_v0_1
        ON connection_v0_1 (item_id);
        """,
        """
        CREATE INDEX IF NOT EXISTS idx_connection_state_v0_1
        ON connection_v0_1 (state);
        """,
        """
        CREATE INDEX IF NOT EXISTS idx_connection_created_at_v0_1
        ON connection_v0_1 (created_at);
        """,
        """
        CREATE OR REPLACE FUNCTION update_connection_timestamp_v0_1()
        RETURNS TRIGGER AS $$
        BEGIN
            IF NEW.updated_at IS NULL THEN
                NEW.updated_at = CURRENT_TIMESTAMP;
            END IF;
            RETURN NEW;
        END;
        $$ LANGUAGE plpgsql;
        """,
        """
        CREATE TRIGGER trg_update_connection_timestamp_v0_1
        BEFORE UPDATE ON connection_v0_1
        FOR EACH ROW
        EXECUTE FUNCTION update_connection_timestamp_v0_1();
        """,
    ],
    "mssql": [
        """
        CREATE TABLE connection_v0_1 (
            id INT IDENTITY(1,1) PRIMARY KEY,
            item_id INT NOT NULL,
            item_name NVARCHAR(MAX),
            state NVARCHAR(255),
            created_at DATETIME2 DEFAULT SYSDATETIME(),
            updated_at DATETIME2 DEFAULT SYSDATETIME(),
            my_did NVARCHAR(255),
            their_did NVARCHAR(255),
            their_label NVARCHAR(MAX),
            their_role NVARCHAR(255),
            invitation_key NVARCHAR(255),
            invitation_msg_id NVARCHAR(255),
            request_id NVARCHAR(255),
            inbound_connection_id NVARCHAR(255),
            error_msg NVARCHAR(MAX),
            accept NVARCHAR(255),
            invitation_mode NVARCHAR(255),
            alias NVARCHAR(MAX),
            their_public_did NVARCHAR(255),
            connection_protocol NVARCHAR(255),
            rfc23_state NVARCHAR(255),
            CONSTRAINT fk_item_id FOREIGN KEY (item_id)
            REFERENCES items(id) ON DELETE CASCADE ON UPDATE CASCADE
        );
        """,
        """
        CREATE NONCLUSTERED INDEX idx_connection_item_id_v0_1
        ON connection_v0_1 (item_id);
        """,
        """
        CREATE NONCLUSTERED INDEX idx_connection_state_v0_1
        ON connection_v0_1 (state);
        """,
        """
        CREATE NONCLUSTERED INDEX idx_connection_created_at_v0_1
        ON connection_v0_1 (created_at);
        """,
        """
        CREATE TRIGGER trg_update_connection_timestamp_v0_1
        ON connection_v0_1
        AFTER UPDATE
        AS
        BEGIN
            UPDATE connection_v0_1
            SET updated_at = SYSDATETIME()
            FROM connection_v0_1
            INNER JOIN inserted ON connection_v0_1.id = inserted.id
            WHERE inserted.updated_at IS NULL;
        END;
        """,
    ],
}

DROP_SCHEMAS = {
    "sqlite": [
        "DROP TRIGGER IF EXISTS trg_update_connection_timestamp_v0_1;",
        "DROP INDEX IF EXISTS idx_connection_created_at_v0_1;",
        "DROP INDEX IF EXISTS idx_connection_state_v0_1;",
        "DROP INDEX IF EXISTS idx_connection_item_id_v0_1;",
        "DROP TABLE IF EXISTS connection_v0_1;",
    ],
    "postgresql": [
        "DROP TRIGGER IF EXISTS trg_update_connection_timestamp_v0_1 ON connection_v0_1;",
        "DROP FUNCTION IF EXISTS update_connection_timestamp_v0_1 CASCADE;",
        "DROP INDEX IF EXISTS idx_connection_created_at_v0_1;",
        "DROP INDEX IF EXISTS idx_connection_state_v0_1;",
        "DROP INDEX IF EXISTS idx_connection_item_id_v0_1;",
        "DROP TABLE IF EXISTS connection_v0_1 CASCADE;",
    ],
    "mssql": [
        "DROP TRIGGER IF EXISTS trg_update_connection_timestamp_v0_1;",
        "DROP INDEX IF EXISTS idx_connection_created_at_v0_1 ON connection_v0_1;",
        "DROP INDEX IF EXISTS idx_connection_state_v0_1 ON connection_v0_1;",
        "DROP INDEX IF EXISTS idx_connection_item_id_v0_1 ON connection_v0_1;",
        "DROP TABLE IF EXISTS connection_v0_1;",
    ],
}

COLUMNS = [
    "state",
    "created_at",
    "updated_at",
    "my_did",
    "their_did",
    "their_label",
    "their_role",
    "invitation_key",
    "invitation_msg_id",
    "request_id",
    "inbound_connection_id",
    "error_msg",
    "accept",
    "invitation_mode",
    "alias",
    "their_public_did",
    "connection_protocol",
    "rfc23_state",
]



"""Module docstring."""

from abc import ABC, abstractmethod
from typing import Generator, Optional, Sequence

from .db_types import Entry


class DatabaseBackend(ABC):
    """Abstract base class for database backends."""

    @abstractmethod
    def provision(
        self,
        uri,
        key_method,
        pass_key,
        profile,
        recreate,
        release_number: str = "release_0",
    ):
        """Provision a new database with the specified release number.

        Args:
            uri: The database URI.
            key_method: Optional key method for encryption.
            pass_key: Optional encryption key.
            profile: Optional profile name.
            recreate: If True, recreate the database.
            release_number: Release number to use (e.g., 'release_0').
                Defaults to 'release_0'.

        """
        pass

    @abstractmethod
    def open(self, uri, key_method, pass_key, profile, release_number: str = "release_0"):
        """Open an existing database with the specified release number.

        Args:
            uri: The database URI.
            key_method: Optional key method for encryption.
            pass_key: Optional encryption key.
            profile: Optional profile name.
            release_number: Release number to use (e.g., 'release_0').
                Defaults to 'release_0'.

        """
        pass

    @abstractmethod
    def remove(self, uri, release_number: str = "release_0"):
        """Remove the database.

        Args:
            uri: The database URI.
            release_number: Release number to use (e.g., 'release_0').
                Defaults to 'release_0'.

        """
        pass

    @abstractmethod
    def translate_error(self, exception):
        """Translate backend-specific exceptions to DBStoreError."""
        pass


class AbstractDatabaseStore(ABC):
    """Abstract base class for database store implementations."""

    @abstractmethod
    async def create_profile(self, name: str = None) -> str:
        """Create a new profile."""
        pass

    @abstractmethod
    async def get_profile_name(self) -> str:
        """Get the profile name."""
        pass

    @abstractmethod
    async def remove_profile(self, name: str) -> bool:
        """Remove a profile."""
        pass

    @abstractmethod
    async def rekey(self, key_method: str = None, pass_key: str = None):
        """Re-key the database."""
        pass

    @abstractmethod
    def scan(
        self,
        profile: Optional[str],
        category: str,
        tag_filter: str | dict = None,
        offset: int = None,
        limit: int = None,
    ) -> Generator[Entry, None, None]:
        """Scan database entries."""
        pass

    @abstractmethod
    def session(
        self, profile: str = None, release_number: str = "release_0"
    ) -> "AbstractDatabaseSession":
        """Create a new database session with the specified release number.

        Args:
            profile: Optional profile name.
            release_number: Release number to use (e.g., 'release_0').
                Defaults to 'release_0'.

        Returns:
            AbstractDatabaseSession: The session instance.

        """
        pass

    @abstractmethod
    def transaction(
        self, profile: str = None, release_number: str = "release_0"
    ) -> "AbstractDatabaseSession":
        """Create a new database transaction with the specified release number.

        Args:
            profile: Optional profile name.
            release_number: Release number to use (e.g., 'release_0').
                Defaults to 'release_0'.

        Returns:
            AbstractDatabaseSession: The transaction instance.

        """
        pass

    @abstractmethod
    async def close(self, remove: bool = False) -> bool:
        """Close the database store."""
        pass


class AbstractDatabaseSession(ABC):
    """Abstract base class for database session implementations."""

    @abstractmethod
    async def count(self, category: str, tag_filter: str | dict = None) -> int:
        """Count entries."""
        pass

    @abstractmethod
    async def fetch(
        self, category: str, name: str, for_update: bool = False
    ) -> Optional[Entry]:
        """Fetch a single entry."""
        pass

    @abstractmethod
    async def fetch_all(
        self,
        category: str,
        tag_filter: str | dict = None,
        limit: int = None,
        for_update: bool = False,
    ) -> Sequence[Entry]:
        """Fetch all matching entries."""
        pass

    @abstractmethod
    async def insert(
        self,
        category: str,
        name: str,
        value: str | bytes = None,
        tags: dict = None,
        expiry_ms: int = None,
        value_json=None,
    ):
        """Insert a new entry."""
        pass

    @abstractmethod
    async def replace(
        self,
        category: str,
        name: str,
        value: str | bytes = None,
        tags: dict = None,
        expiry_ms: int = None,
        value_json=None,
    ):
        """Replace an existing entry."""
        pass

    @abstractmethod
    async def remove(self, category: str, name: str):
        """Remove an entry."""
        pass

    @abstractmethod
    async def remove_all(self, category: str, tag_filter: str | dict = None) -> int:
        """Remove all matching entries."""
        pass

    @abstractmethod
    async def commit(self):
        """Commit the transaction."""
        pass

    @abstractmethod
    async def rollback(self):
        """Rollback the transaction."""
        pass

    @abstractmethod
    async def close(self):
        """Close the session."""
        pass



# acapy_agent/database_manager/databases/sqlite/__init__.py


from ..errors import DatabaseError, DatabaseErrorCode
from .config import SqliteConfig
from .connection_pool import ConnectionPool

__all__ = ["ConnectionPool", "SqliteConfig", "DatabaseErrorCode", "DatabaseError"]



"""AnonCreds schema models."""

from marshmallow import fields

from ....messaging.models.openapi import OpenAPISchema
from ....messaging.valid import (
    ANONCREDS_DID_EXAMPLE,
    ANONCREDS_SCHEMA_ID_EXAMPLE,
)
from ...models.schema import AnonCredsSchemaSchema
from ..common.schemas import EndorserOptionsSchema, SchemaQueryFieldsMixin


class SchemaIdMatchInfo(OpenAPISchema):
    """Path parameters and validators for request taking schema id."""

    schema_id = fields.Str(
        metadata={
            "description": "Schema identifier",
            "example": ANONCREDS_SCHEMA_ID_EXAMPLE,
        }
    )


class SchemaPostOptionSchema(EndorserOptionsSchema):
    """Parameters and validators for schema options."""

    pass


class SchemasQueryStringSchema(SchemaQueryFieldsMixin):
    """Parameters and validators for query string in schemas list query."""

    schema_issuer_id = fields.Str(
        metadata={
            "description": "Schema issuer identifier",
            "example": ANONCREDS_DID_EXAMPLE,
        }
    )


class GetSchemasResponseSchema(OpenAPISchema):
    """Parameters and validators for schema list all response."""

    schema_ids = fields.List(
        fields.Str(
            metadata={
                "description": "Schema identifiers",
                "example": ANONCREDS_SCHEMA_ID_EXAMPLE,
            }
        )
    )


class SchemaPostRequestSchema(OpenAPISchema):
    """Parameters and validators for query string in create schema."""

    schema = fields.Nested(AnonCredsSchemaSchema())
    options = fields.Nested(SchemaPostOptionSchema())



# pytest --maxfail=1 --disable-warnings --no-cov -s -vv acapy_agent/database_manager/wql_normalized/tests/test_postgres_TagsqlEncoder_All_key_value_A.py
# python -m unittest acapy_agent.database_manager.wql_normalized.tests.test_postgres_TagsqlEncoder_All_key_value_A -v

import logging
import os
import unittest

import psycopg
import pytest

from acapy_agent.database_manager.wql_normalized.encoders import encoder_factory
from acapy_agent.database_manager.wql_normalized.query import query_from_str
from acapy_agent.database_manager.wql_normalized.tags import (
    TagName,
    TagQuery,
    query_to_tagquery,
)

# Set up logging
logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


def replace_placeholders(query, args):
    """Replace each '%s' in the query with the corresponding argument for logging.

    Properly quote arguments for SQL, escaping single quotes by doubling them.
    Example: 'O'Reilly' becomes 'O''Reilly'.
    """
    result = query
    for arg in args:
        escaped_arg = str(arg).replace("'", "''")  # Escape single quotes for SQL
        result = result.replace("%s", f"'{escaped_arg}'", 1)  # Replace one %s at a time
    return result


@pytest.mark.postgres
class TestPostgresTagEncoderNonNormalizedA(unittest.TestCase):
    """Test cases for the PostgresTagEncoder class in non-normalized mode (part A)."""

    def setUp(self):
        """Set up PostgreSQL database connection and encoder."""
        self.enc_name = lambda x: x  # No transformation for tag names
        self.enc_value = lambda x: x  # No transformation for tag values

        # Get PostgreSQL connection from environment variable or use default
        postgres_url = os.environ.get(
            "POSTGRES_URL", "postgres://myuser:mypass@localhost:5432/mydb2"
        )
        # Parse the URL to extract connection parameters
        import urllib.parse

        parsed = urllib.parse.urlparse(postgres_url)

        try:
            self.conn = psycopg.connect(
                host=parsed.hostname or "localhost",
                port=parsed.port or 5432,
                dbname=parsed.path.lstrip("/") if parsed.path else "mydb2",
                user=parsed.username or "myuser",
                password=parsed.password or "mypass",
            )
            self.conn.autocommit = True  # Enable autocommit for setup/teardown
            self.cursor = self.conn.cursor()
            # Create tables for key-value pair structure
            self.cursor.execute("""
                CREATE TABLE IF NOT EXISTS items (
                    id SERIAL PRIMARY KEY
                )
            """)
            self.cursor.execute("""
                CREATE TABLE IF NOT EXISTS items_tags (
                    item_id INTEGER,
                    name TEXT,
                    value TEXT,
                    FOREIGN KEY(item_id) REFERENCES items(id)
                )
            """)
            logger.info("Tables 'items' and 'items_tags' created in setUp")
            self.encoder = encoder_factory.get_encoder(
                "postgresql", self.enc_name, self.enc_value, normalized=False
            )
        except Exception as e:
            logger.error(f"Failed to set up PostgreSQL database: {e}")
            raise

    def tearDown(self):
        """Clean up by dropping tables and closing the PostgreSQL connection."""
        try:
            self.cursor.execute("DROP TABLE IF EXISTS items_tags")
            self.cursor.execute("DROP TABLE IF EXISTS items")
            self.conn.commit()
            self.cursor.close()
            self.conn.close()
            logger.info("Tables dropped and PostgreSQL connection closed in tearDown")
        except Exception as e:
            logger.error(f"Failed to tear down PostgreSQL connection: {e}")
            raise

    def run_query_and_verify(self, sql_query, params, expected_ids, test_name):
        """Run a PostgreSQL query and verify results."""
        try:
            query = sql_query[0] if isinstance(sql_query, tuple) else sql_query
            self.cursor.execute(f"SELECT i.id FROM items i WHERE {query}", params)
            actual_ids = sorted([row[0] for row in self.cursor.fetchall()])
            self.assertEqual(
                actual_ids,
                expected_ids,
                f"{test_name} failed: Expected IDs {expected_ids}, got {actual_ids}",
            )
        except Exception as e:
            logger.error(f"Query execution failed in {test_name}: {e}")
            raise

    def verify_round_trip(self, query, original_sql_query, original_params):
        """Verify that converting TagQuery to WQL and back results in the same PostgreSQL query."""
        wql_str = query.to_wql_str()
        parsed_query = query_from_str(wql_str)
        parsed_tag_query = query_to_tagquery(parsed_query)
        parsed_sql_query, parsed_params = self.encoder.encode_query(parsed_tag_query)
        self.assertEqual(
            (original_sql_query, original_params),
            (parsed_sql_query, parsed_params),
            f"Round-trip PostgreSQL query mismatch in {self._testMethodName}",
        )

    def test_eq_positive(self):
        query = TagQuery.eq(TagName("field"), "value")
        wql = query.to_wql_str()
        print(f"Test: Positive equality query\nWQL: {wql}")
        sql_query, params = self.encoder.encode_query(query)
        expected_query = (
            "i.id IN (SELECT item_id FROM items_tags WHERE name = %s AND value = %s)"
        )
        expected_params = ["field", "value"]
        self.assertEqual(sql_query, expected_query, "Positive equality query mismatch")
        self.assertEqual(params, expected_params, "Positive equality params mismatch")
        self.verify_round_trip(query, sql_query, params)
        self.cursor.executemany(
            "INSERT INTO items (id) VALUES (%s) RETURNING id", [(1,), (2,), (3,)]
        )
        self.cursor.executemany(
            "INSERT INTO items_tags (item_id, name, value) VALUES (%s, %s, %s)",
            [(1, "field", "value"), (2, "field", "other"), (3, "field", "value")],
        )
        self.conn.commit()
        print("\n### Complete SQL Statements for Testing")
        print("CREATE TABLE items (id SERIAL PRIMARY KEY);")
        print(
            "CREATE TABLE items_tags (item_id INTEGER, name TEXT, value TEXT, FOREIGN KEY(item_id) REFERENCES items(id));"
        )
        print("INSERT INTO items (id) VALUES (1), (2), (3);")
        print(
            "INSERT INTO items_tags (item_id, name, value) VALUES "
            "(1, 'field', 'value'), "
            "(2, 'field', 'other'), "
            "(3, 'field', 'value');"
        )
        select_query = f"SELECT id FROM items i WHERE {sql_query}"
        complete_sql = replace_placeholders(select_query, params)
        print(f"\n-- Complete SELECT statement with values:\n{complete_sql}")
        print("\n-- Expected result: Items 1, 3")
        print("\n-- Cleanup\nDROP TABLE items_tags; DROP TABLE items;")
        self.run_query_and_verify(sql_query, params, [1, 3], "Positive equality")

    def test_eq_negated(self):
        query = TagQuery.not_(TagQuery.eq(TagName("field"), "value"))
        wql = query.to_wql_str()
        print(f"Test: Negated equality query\nWQL: {wql}")
        sql_query, params = self.encoder.encode_query(query)
        expected_query = (
            "i.id NOT IN (SELECT item_id FROM items_tags WHERE name = %s AND value = %s)"
        )
        expected_params = ["field", "value"]
        self.assertEqual(sql_query, expected_query, "Negated equality query mismatch")
        self.assertEqual(params, expected_params, "Negated equality params mismatch")
        self.verify_round_trip(query, sql_query, params)
        self.cursor.executemany(
            "INSERT INTO items (id) VALUES (%s) RETURNING id", [(1,), (2,), (3,)]
        )
        self.cursor.executemany(
            "INSERT INTO items_tags (item_id, name, value) VALUES (%s, %s, %s)",
            [(1, "field", "value"), (2, "field", "other"), (3, "field", "value")],
        )
        self.conn.commit()
        print("\n### Complete SQL Statements for Testing")
        print("CREATE TABLE items (id SERIAL PRIMARY KEY);")
        print(
            "CREATE TABLE items_tags (item_id INTEGER, name TEXT, value TEXT, FOREIGN KEY(item_id) REFERENCES items(id));"
        )
        print("INSERT INTO items (id) VALUES (1), (2), (3);")
        print(
            "INSERT INTO items_tags (item_id, name, value) VALUES "
            "(1, 'field', 'value'), "
            "(2, 'field', 'other'), "
            "(3, 'field', 'value');"
        )
        select_query = f"SELECT id FROM items i WHERE {sql_query}"
        complete_sql = replace_placeholders(select_query, params)
        print(f"\n-- Complete SELECT statement with values:\n{complete_sql}")
        print("\n-- Expected result: Item 2")
        print("\n-- Cleanup\nDROP TABLE items_tags; DROP TABLE items;")
        self.run_query_and_verify(sql_query, params, [2], "Negated equality")

    def test_neq_positive(self):
        query = TagQuery.neq(TagName("field"), "value")
        wql = query.to_wql_str()
        print(f"Test: Positive inequality query\nWQL: {wql}")
        sql_query, params = self.encoder.encode_query(query)
        expected_query = (
            "i.id IN (SELECT item_id FROM items_tags WHERE name = %s AND value != %s)"
        )
        expected_params = ["field", "value"]
        self.assertEqual(sql_query, expected_query, "Positive inequality query mismatch")
        self.assertEqual(params, expected_params, "Positive inequality params mismatch")
        self.verify_round_trip(query, sql_query, params)
        self.cursor.executemany(
            "INSERT INTO items (id) VALUES (%s) RETURNING id", [(1,), (2,), (3,)]
        )
        self.cursor.executemany(
            "INSERT INTO items_tags (item_id, name, value) VALUES (%s, %s, %s)",
            [(1, "field", "value"), (2, "field", "other"), (3, "field", "different")],
        )
        self.conn.commit()
        print("\n### Complete SQL Statements for Testing")
        print("CREATE TABLE items (id SERIAL PRIMARY KEY);")
        print(
            "CREATE TABLE items_tags (item_id INTEGER, name TEXT, value TEXT, FOREIGN KEY(item_id) REFERENCES items(id));"
        )
        print("INSERT INTO items (id) VALUES (1), (2), (3);")
        print(
            "INSERT INTO items_tags (item_id, name, value) VALUES "
            "(1, 'field', 'value'), "
            "(2, 'field', 'other'), "
            "(3, 'field', 'different');"
        )
        select_query = f"SELECT id FROM items i WHERE {sql_query}"
        complete_sql = replace_placeholders(select_query, params)
        print(f"\n-- Complete SELECT statement with values:\n{complete_sql}")
        print("\n-- Expected result: Items 2, 3")
        print("\n-- Cleanup\nDROP TABLE items_tags; DROP TABLE items;")
        self.run_query_and_verify(sql_query, params, [2, 3], "Positive inequality")

    def test_neq_negated(self):
        query = TagQuery.not_(TagQuery.neq(TagName("field"), "value"))
        wql = query.to_wql_str()
        print(f"Test: Negated inequality query\nWQL: {wql}")
        sql_query, params = self.encoder.encode_query(query)
        expected_query = (
            "i.id NOT IN (SELECT item_id FROM items_tags WHERE name = %s AND value != %s)"
        )
        expected_params = ["field", "value"]
        self.assertEqual(sql_query, expected_query, "Negated inequality query mismatch")
        self.assertEqual(params, expected_params, "Negated inequality params mismatch")
        self.verify_round_trip(query, sql_query, params)
        self.cursor.executemany(
            "INSERT INTO items (id) VALUES (%s) RETURNING id", [(1,), (2,), (3,)]
        )
        self.cursor.executemany(
            "INSERT INTO items_tags (item_id, name, value) VALUES (%s, %s, %s)",
            [(1, "field", "value"), (2, "field", "other"), (3, "field", "value")],
        )
        self.conn.commit()
        print("\n### Complete SQL Statements for Testing")
        print("CREATE TABLE items (id SERIAL PRIMARY KEY);")
        print(
            "CREATE TABLE items_tags (item_id INTEGER, name TEXT, value TEXT, FOREIGN KEY(item_id) REFERENCES items(id));"
        )
        print("INSERT INTO items (id) VALUES (1), (2), (3);")
        print(
            "INSERT INTO items_tags (item_id, name, value) VALUES "
            "(1, 'field', 'value'), "
            "(2, 'field', 'other'), "
            "(3, 'field', 'value');"
        )
        select_query = f"SELECT id FROM items i WHERE {sql_query}"
        complete_sql = replace_placeholders(select_query, params)
        print(f"\n-- Complete SELECT statement with values:\n{complete_sql}")
        print("\n-- Expected result: Items 1, 3")
        print("\n-- Cleanup\nDROP TABLE items_tags; DROP TABLE items;")
        self.run_query_and_verify(sql_query, params, [1, 3], "Negated inequality")

    def test_gt_positive(self):
        query = TagQuery.gt(TagName("price"), "100")
        wql = query.to_wql_str()
        print(f"Test: Positive greater-than query\nWQL: {wql}")
        sql_query, params = self.encoder.encode_query(query)
        expected_query = (
            "i.id IN (SELECT item_id FROM items_tags WHERE name = %s AND value > %s)"
        )
        expected_params = ["price", "100"]
        self.assertEqual(
            sql_query, expected_query, "Positive greater-than query mismatch"
        )
        self.assertEqual(params, expected_params, "Positive greater-than params mismatch")
        self.verify_round_trip(query, sql_query, params)
        self.cursor.executemany(
            "INSERT INTO items (id) VALUES (%s) RETURNING id", [(1,), (2,), (3,), (4,)]
        )
        self.cursor.executemany(
            "INSERT INTO items_tags (item_id, name, value) VALUES (%s, %s, %s)",
            [
                (1, "price", "090"),
                (2, "price", "100"),
                (3, "price", "150"),
                (4, "price", "200"),
            ],
        )
        self.conn.commit()
        print("\n### Complete SQL Statements for Testing")
        print("CREATE TABLE items (id SERIAL PRIMARY KEY);")
        print(
            "CREATE TABLE items_tags (item_id INTEGER, name TEXT, value TEXT, FOREIGN KEY(item_id) REFERENCES items(id));"
        )
        print("INSERT INTO items (id) VALUES (1), (2), (3), (4);")
        print(
            "INSERT INTO items_tags (item_id, name, value) VALUES "
            "(1, 'price', '090'), "
            "(2, 'price', '100'), "
            "(3, 'price', '150'), "
            "(4, 'price', '200');"
        )
        select_query = f"SELECT id FROM items i WHERE {sql_query}"
        complete_sql = replace_placeholders(select_query, params)
        print(f"\n-- Complete SELECT statement with values:\n{complete_sql}")
        print("\n-- Expected result: Items 3, 4")
        print("\n-- Cleanup\nDROP TABLE items_tags; DROP TABLE items;")
        self.run_query_and_verify(sql_query, params, [3, 4], "Positive greater-than")

    def test_gt_negated(self):
        query = TagQuery.not_(TagQuery.gt(TagName("price"), "100"))
        wql = query.to_wql_str()
        print(f"Test: Negated greater-than query\nWQL: {wql}")
        sql_query, params = self.encoder.encode_query(query)
        expected_query = (
            "i.id NOT IN (SELECT item_id FROM items_tags WHERE name = %s AND value > %s)"
        )
        expected_params = ["price", "100"]
        self.assertEqual(sql_query, expected_query, "Negated greater-than query mismatch")
        self.assertEqual(params, expected_params, "Negated greater-than params mismatch")
        self.verify_round_trip(query, sql_query, params)
        self.cursor.executemany(
            "INSERT INTO items (id) VALUES (%s) RETURNING id", [(1,), (2,), (3,), (4,)]
        )
        self.cursor.executemany(
            "INSERT INTO items_tags (item_id, name, value) VALUES (%s, %s, %s)",
            [
                (1, "price", "090"),
                (2, "price", "100"),
                (3, "price", "150"),
                (4, "price", "200"),
            ],
        )
        self.conn.commit()
        print("\n### Complete SQL Statements for Testing")
        print("CREATE TABLE items (id SERIAL PRIMARY KEY);")
        print(
            "CREATE TABLE items_tags (item_id INTEGER, name TEXT, value TEXT, FOREIGN KEY(item_id) REFERENCES items(id));"
        )
        print("INSERT INTO items (id) VALUES (1), (2), (3), (4);")
        print(
            "INSERT INTO items_tags (item_id, name, value) VALUES "
            "(1, 'price', '090'), "
            "(2, 'price', '100'), "
            "(3, 'price', '150'), "
            "(4, 'price', '200');"
        )
        select_query = f"SELECT id FROM items i WHERE {sql_query}"
        complete_sql = replace_placeholders(select_query, params)
        print(f"\n-- Complete SELECT statement with values:\n{complete_sql}")
        print("\n-- Expected result: Items 1, 2")
        print("\n-- Cleanup\nDROP TABLE items_tags; DROP TABLE items;")
        self.run_query_and_verify(sql_query, params, [1, 2], "Negated greater-than")

    def test_gte_positive(self):
        query = TagQuery.gte(TagName("price"), "100")
        wql = query.to_wql_str()
        print(f"Test: Positive greater-than-or-equal query\nWQL: {wql}")
        sql_query, params = self.encoder.encode_query(query)
        expected_query = (
            "i.id IN (SELECT item_id FROM items_tags WHERE name = %s AND value >= %s)"
        )
        expected_params = ["price", "100"]
        self.assertEqual(
            sql_query, expected_query, "Positive greater-than-or-equal query mismatch"
        )
        self.assertEqual(
            params, expected_params, "Positive greater-than-or-equal params mismatch"
        )
        self.verify_round_trip(query, sql_query, params)
        self.cursor.executemany(
            "INSERT INTO items (id) VALUES (%s) RETURNING id", [(1,), (2,), (3,), (4,)]
        )
        self.cursor.executemany(
            "INSERT INTO items_tags (item_id, name, value) VALUES (%s, %s, %s)",
            [
                (1, "price", "090"),
                (2, "price", "100"),
                (3, "price", "150"),
                (4, "price", "200"),
            ],
        )
        self.conn.commit()
        print("\n### Complete SQL Statements for Testing")
        print("CREATE TABLE items (id SERIAL PRIMARY KEY);")
        print(
            "CREATE TABLE items_tags (item_id INTEGER, name TEXT, value TEXT, FOREIGN KEY(item_id) REFERENCES items(id));"
        )
        print("INSERT INTO items (id) VALUES (1), (2), (3), (4);")
        print(
            "INSERT INTO items_tags (item_id, name, value) VALUES "
            "(1, 'price', '090'), "
            "(2, 'price', '100'), "
            "(3, 'price', '150'), "
            "(4, 'price', '200');"
        )
        select_query = f"SELECT id FROM items i WHERE {sql_query}"
        complete_sql = replace_placeholders(select_query, params)
        print(f"\n-- Complete SELECT statement with values:\n{complete_sql}")
        print("\n-- Expected result: Items 2, 3, 4")
        print("\n-- Cleanup\nDROP TABLE items_tags; DROP TABLE items;")
        self.run_query_and_verify(
            sql_query, params, [2, 3, 4], "Positive greater-than-or-equal"
        )

    def test_gte_negated(self):
        query = TagQuery.not_(TagQuery.gte(TagName("price"), "100"))
        wql = query.to_wql_str()
        print(f"Test: Negated greater-than-or-equal query\nWQL: {wql}")
        sql_query, params = self.encoder.encode_query(query)
        expected_query = (
            "i.id NOT IN (SELECT item_id FROM items_tags WHERE name = %s AND value >= %s)"
        )
        expected_params = ["price", "100"]
        self.assertEqual(
            sql_query, expected_query, "Negated greater-than-or-equal query mismatch"
        )
        self.assertEqual(
            params, expected_params, "Negated greater-than-or-equal params mismatch"
        )
        self.verify_round_trip(query, sql_query, params)
        self.cursor.executemany(
            "INSERT INTO items (id) VALUES (%s) RETURNING id", [(1,), (2,), (3,), (4,)]
        )
        self.cursor.executemany(
            "INSERT INTO items_tags (item_id, name, value) VALUES (%s, %s, %s)",
            [
                (1, "price", "090"),
                (2, "price", "100"),
                (3, "price", "150"),
                (4, "price", "200"),
            ],
        )
        self.conn.commit()
        print("\n### Complete SQL Statements for Testing")
        print("CREATE TABLE items (id SERIAL PRIMARY KEY);")
        print(
            "CREATE TABLE items_tags (item_id INTEGER, name TEXT, value TEXT, FOREIGN KEY(item_id) REFERENCES items(id));"
        )
        print("INSERT INTO items (id) VALUES (1), (2), (3), (4);")
        print(
            "INSERT INTO items_tags (item_id, name, value) VALUES "
            "(1, 'price', '090'), "
            "(2, 'price', '100'), "
            "(3, 'price', '150'), "
            "(4, 'price', '200');"
        )
        select_query = f"SELECT id FROM items i WHERE {sql_query}"
        complete_sql = replace_placeholders(select_query, params)
        print(f"\n-- Complete SELECT statement with values:\n{complete_sql}")
        print("\n-- Expected result: Item 1")
        print("\n-- Cleanup\nDROP TABLE items_tags; DROP TABLE items;")
        self.run_query_and_verify(sql_query, params, [1], "Negated greater-than-or-equal")

    def test_lt_positive(self):
        query = TagQuery.lt(TagName("price"), "100")
        wql = query.to_wql_str()
        print(f"Test: Positive less-than query\nWQL: {wql}")
        sql_query, params = self.encoder.encode_query(query)
        expected_query = (
            "i.id IN (SELECT item_id FROM items_tags WHERE name = %s AND value < %s)"
        )
        expected_params = ["price", "100"]
        self.assertEqual(sql_query, expected_query, "Positive less-than query mismatch")
        self.assertEqual(params, expected_params, "Positive less-than params mismatch")
        self.verify_round_trip(query, sql_query, params)
        self.cursor.executemany(
            "INSERT INTO items (id) VALUES (%s) RETURNING id", [(1,), (2,), (3,), (4,)]
        )
        self.cursor.executemany(
            "INSERT INTO items_tags (item_id, name, value) VALUES (%s, %s, %s)",
            [
                (1, "price", "090"),
                (2, "price", "100"),
                (3, "price", "150"),
                (4, "price", "200"),
            ],
        )
        self.conn.commit()
        print("\n### Complete SQL Statements for Testing")
        print("CREATE TABLE items (id SERIAL PRIMARY KEY);")
        print(
            "CREATE TABLE items_tags (item_id INTEGER, name TEXT, value TEXT, FOREIGN KEY(item_id) REFERENCES items(id));"
        )
        print("INSERT INTO items (id) VALUES (1), (2), (3), (4);")
        print(
            "INSERT INTO items_tags (item_id, name, value) VALUES "
            "(1, 'price', '090'), "
            "(2, 'price', '100'), "
            "(3, 'price', '150'), "
            "(4, 'price', '200');"
        )
        select_query = f"SELECT id FROM items i WHERE {sql_query}"
        complete_sql = replace_placeholders(select_query, params)
        print(f"\n-- Complete SELECT statement with values:\n{complete_sql}")
        print("\n-- Expected result: Item 1")
        print("\n-- Cleanup\nDROP TABLE items_tags; DROP TABLE items;")
        self.run_query_and_verify(sql_query, params, [1], "Positive less-than")

    def test_lt_negated(self):
        query = TagQuery.not_(TagQuery.lt(TagName("price"), "100"))
        wql = query.to_wql_str()
        print(f"Test: Negated less-than query\nWQL: {wql}")
        sql_query, params = self.encoder.encode_query(query)
        expected_query = (
            "i.id NOT IN (SELECT item_id FROM items_tags WHERE name = %s AND value < %s)"
        )
        expected_params = ["price", "100"]
        self.assertEqual(sql_query, expected_query, "Negated less-than query mismatch")
        self.assertEqual(params, expected_params, "Negated less-than params mismatch")
        self.verify_round_trip(query, sql_query, params)
        self.cursor.executemany(
            "INSERT INTO items (id) VALUES (%s) RETURNING id", [(1,), (2,), (3,), (4,)]
        )
        self.cursor.executemany(
            "INSERT INTO items_tags (item_id, name, value) VALUES (%s, %s, %s)",
            [
                (1, "price", "090"),
                (2, "price", "100"),
                (3, "price", "150"),
                (4, "price", "200"),
            ],
        )
        self.conn.commit()
        print("\n### Complete SQL Statements for Testing")
        print("CREATE TABLE items (id SERIAL PRIMARY KEY);")
        print(
            "CREATE TABLE items_tags (item_id INTEGER, name TEXT, value TEXT, FOREIGN KEY(item_id) REFERENCES items(id));"
        )
        print("INSERT INTO items (id) VALUES (1), (2), (3), (4);")
        print(
            "INSERT INTO items_tags (item_id, name, value) VALUES "
            "(1, 'price', '090'), "
            "(2, 'price', '100'), "
            "(3, 'price', '150'), "
            "(4, 'price', '200');"
        )
        select_query = f"SELECT id FROM items i WHERE {sql_query}"
        complete_sql = replace_placeholders(select_query, params)
        print(f"\n-- Complete SELECT statement with values:\n{complete_sql}")
        print("\n-- Expected result: Items 2, 3, 4")
        print("\n-- Cleanup\nDROP TABLE items_tags; DROP TABLE items;")
        self.run_query_and_verify(sql_query, params, [2, 3, 4], "Negated less-than")

    def test_lte_positive(self):
        query = TagQuery.lte(TagName("price"), "100")
        wql = query.to_wql_str()
        print(f"Test: Positive less-than-or-equal query\nWQL: {wql}")
        sql_query, params = self.encoder.encode_query(query)
        expected_query = (
            "i.id IN (SELECT item_id FROM items_tags WHERE name = %s AND value <= %s)"
        )
        expected_params = ["price", "100"]
        self.assertEqual(
            sql_query, expected_query, "Positive less-than-or-equal query mismatch"
        )
        self.assertEqual(
            params, expected_params, "Positive less-than-or-equal params mismatch"
        )
        self.verify_round_trip(query, sql_query, params)
        self.cursor.executemany(
            "INSERT INTO items (id) VALUES (%s) RETURNING id", [(1,), (2,), (3,), (4,)]
        )
        self.cursor.executemany(
            "INSERT INTO items_tags (item_id, name, value) VALUES (%s, %s, %s)",
            [
                (1, "price", "090"),
                (2, "price", "100"),
                (3, "price", "150"),
                (4, "price", "200"),
            ],
        )
        self.conn.commit()
        print("\n### Complete SQL Statements for Testing")
        print("CREATE TABLE items (id SERIAL PRIMARY KEY);")
        print(
            "CREATE TABLE items_tags (item_id INTEGER, name TEXT, value TEXT, FOREIGN KEY(item_id) REFERENCES items(id));"
        )
        print("INSERT INTO items (id) VALUES (1), (2), (3), (4);")
        print(
            "INSERT INTO items_tags (item_id, name, value) VALUES "
            "(1, 'price', '090'), "
            "(2, 'price', '100'), "
            "(3, 'price', '150'), "
            "(4, 'price', '200');"
        )
        select_query = f"SELECT id FROM items i WHERE {sql_query}"
        complete_sql = replace_placeholders(select_query, params)
        print(f"\n-- Complete SELECT statement with values:\n{complete_sql}")
        print("\n-- Expected result: Items 1, 2")
        print("\n-- Cleanup\nDROP TABLE items_tags; DROP TABLE items;")
        self.run_query_and_verify(
            sql_query, params, [1, 2], "Positive less-than-or-equal"
        )

    def test_lte_negated(self):
        query = TagQuery.not_(TagQuery.lte(TagName("price"), "100"))
        wql = query.to_wql_str()
        print(f"Test: Negated less-than-or-equal query\nWQL: {wql}")
        sql_query, params = self.encoder.encode_query(query)
        expected_query = (
            "i.id NOT IN (SELECT item_id FROM items_tags WHERE name = %s AND value <= %s)"
        )
        expected_params = ["price", "100"]
        self.assertEqual(
            sql_query, expected_query, "Negated less-than-or-equal query mismatch"
        )
        self.assertEqual(
            params, expected_params, "Negated less-than-or-equal params mismatch"
        )
        self.verify_round_trip(query, sql_query, params)
        self.cursor.executemany(
            "INSERT INTO items (id) VALUES (%s) RETURNING id", [(1,), (2,), (3,), (4,)]
        )
        self.cursor.executemany(
            "INSERT INTO items_tags (item_id, name, value) VALUES (%s, %s, %s)",
            [
                (1, "price", "090"),
                (2, "price", "100"),
                (3, "price", "150"),
                (4, "price", "200"),
            ],
        )
        self.conn.commit()
        print("\n### Complete SQL Statements for Testing")
        print("CREATE TABLE items (id SERIAL PRIMARY KEY);")
        print(
            "CREATE TABLE items_tags (item_id INTEGER, name TEXT, value TEXT, FOREIGN KEY(item_id) REFERENCES items(id));"
        )
        print("INSERT INTO items (id) VALUES (1), (2), (3), (4);")
        print(
            "INSERT INTO items_tags (item_id, name, value) VALUES "
            "(1, 'price', '090'), "
            "(2, 'price', '100'), "
            "(3, 'price', '150'), "
            "(4, 'price', '200');"
        )
        select_query = f"SELECT id FROM items i WHERE {sql_query}"
        complete_sql = replace_placeholders(select_query, params)
        print(f"\n-- Complete SELECT statement with values:\n{complete_sql}")
        print("\n-- Expected result: Items 3, 4")
        print("\n-- Cleanup\nDROP TABLE items_tags; DROP TABLE items;")
        self.run_query_and_verify(sql_query, params, [3, 4], "Negated less-than-or-equal")

    def test_like_positive(self):
        query = TagQuery.like(TagName("field"), "%pat%")
        wql = query.to_wql_str()
        print(f"Test: Positive LIKE query\nWQL: {wql}")
        sql_query, params = self.encoder.encode_query(query)
        expected_query = (
            "i.id IN (SELECT item_id FROM items_tags WHERE name = %s AND value LIKE %s)"
        )
        expected_params = ["field", "%pat%"]
        self.assertEqual(sql_query, expected_query, "Positive LIKE query mismatch")
        self.assertEqual(params, expected_params, "Positive LIKE params mismatch")
        self.verify_round_trip(query, sql_query, params)
        self.cursor.executemany(
            "INSERT INTO items (id) VALUES (%s) RETURNING id", [(1,), (2,), (3,), (4,)]
        )
        self.cursor.executemany(
            "INSERT INTO items_tags (item_id, name, value) VALUES (%s, %s, %s)",
            [
                (1, "field", "pattern"),
                (2, "field", "path"),
                (3, "field", "other"),
                (4, "field", "pat"),
            ],
        )
        self.conn.commit()
        print("\n### Complete SQL Statements for Testing")
        print("CREATE TABLE items (id SERIAL PRIMARY KEY);")
        print(
            "CREATE TABLE items_tags (item_id INTEGER, name TEXT, value TEXT, FOREIGN KEY(item_id) REFERENCES items(id));"
        )
        print("INSERT INTO items (id) VALUES (1), (2), (3), (4);")
        print(
            "INSERT INTO items_tags (item_id, name, value) VALUES "
            "(1, 'field', 'pattern'), "
            "(2, 'field', 'path'), "
            "(3, 'field', 'other'), "
            "(4, 'field', 'pat');"
        )
        select_query = f"SELECT id FROM items i WHERE {sql_query}"
        complete_sql = replace_placeholders(select_query, params)
        print(f"\n-- Complete SELECT statement with values:\n{complete_sql}")
        print("\n-- Expected result: Items 1, 2, 4")
        print("\n-- Cleanup\nDROP TABLE items_tags; DROP TABLE items;")
        self.run_query_and_verify(sql_query, params, [1, 2, 4], "Positive LIKE")

    def test_like_negated(self):
        query = TagQuery.not_(TagQuery.like(TagName("field"), "%pat%"))
        wql = query.to_wql_str()
        print(f"Test: Negated LIKE query\nWQL: {wql}")
        sql_query, params = self.encoder.encode_query(query)
        expected_query = "i.id NOT IN (SELECT item_id FROM items_tags WHERE name = %s AND value LIKE %s)"
        expected_params = ["field", "%pat%"]
        self.assertEqual(sql_query, expected_query, "Negated LIKE query mismatch")
        self.assertEqual(params, expected_params, "Negated LIKE params mismatch")
        self.verify_round_trip(query, sql_query, params)
        self.cursor.executemany(
            "INSERT INTO items (id) VALUES (%s) RETURNING id", [(1,), (2,), (3,), (4,)]
        )
        self.cursor.executemany(
            "INSERT INTO items_tags (item_id, name, value) VALUES (%s, %s, %s)",
            [
                (1, "field", "pattern"),
                (2, "field", "path"),
                (3, "field", "other"),
                (4, "field", "pat"),
            ],
        )
        self.conn.commit()
        print("\n### Complete SQL Statements for Testing")
        print("CREATE TABLE items (id SERIAL PRIMARY KEY);")
        print(
            "CREATE TABLE items_tags (item_id INTEGER, name TEXT, value TEXT, FOREIGN KEY(item_id) REFERENCES items(id));"
        )
        print("INSERT INTO items (id) VALUES (1), (2), (3), (4);")
        print(
            "INSERT INTO items_tags (item_id, name, value) VALUES "
            "(1, 'field', 'pattern'), "
            "(2, 'field', 'path'), "
            "(3, 'field', 'other'), "
            "(4, 'field', 'pat');"
        )
        select_query = f"SELECT id FROM items i WHERE {sql_query}"
        complete_sql = replace_placeholders(select_query, params)
        print(f"\n-- Complete SELECT statement with values:\n{complete_sql}")
        print("\n-- Expected result: Item 3")
        print("\n-- Cleanup\nDROP TABLE items_tags; DROP TABLE items;")
        self.run_query_and_verify(sql_query, params, [3], "Negated LIKE")


def main():
    print("Running PostgresTagEncoder non-normalized tests (part A)...")
    unittest.main(argv=[""], exit=False)
    print("All tests completed.")


if __name__ == "__main__":
    main()



"""Automated setup process for AnonCreds credential definitions with revocation."""

import asyncio
import logging
from abc import ABC, abstractmethod
from collections.abc import Awaitable, Callable
from typing import Any

from ...core.event_bus import Event, EventBus
from ...core.profile import Profile
from ...revocation.util import notify_revocation_published_event
from ...storage.type import (
    RECORD_TYPE_REV_LIST_CREATE_EVENT,
    RECORD_TYPE_REV_LIST_STORE_EVENT,
    RECORD_TYPE_REV_REG_ACTIVATION_EVENT,
    RECORD_TYPE_REV_REG_DEF_CREATE_EVENT,
    RECORD_TYPE_REV_REG_DEF_STORE_EVENT,
    RECORD_TYPE_REV_REG_FULL_HANDLING_EVENT,
)
from ..events import (
    FIRST_REGISTRY_TAG,
    INTERVENTION_REQUIRED_EVENT,
    BaseEventPayload,
    BasePayloadWithFailure,
    CredDefFinishedEvent,
    InterventionRequiredPayload,
    RevListCreateRequestedEvent,
    RevListCreateResponseEvent,
    RevListFinishedEvent,
    RevListStoreRequestedEvent,
    RevListStoreResponseEvent,
    RevRegActivationRequestedEvent,
    RevRegActivationResponseEvent,
    RevRegDefCreateRequestedEvent,
    RevRegDefCreateResponseEvent,
    RevRegDefStoreRequestedEvent,
    RevRegDefStoreResponseEvent,
    RevRegFullDetectedEvent,
    RevRegFullHandlingResponseEvent,
)
from ..issuer import STATE_FINISHED
from ..revocation import AnonCredsRevocation
from .auto_recovery import (
    EventStorageManager,
    calculate_event_expiry_timestamp,
    calculate_exponential_backoff_delay,
    generate_correlation_id,
    generate_request_id,
    serialize_event_payload,
)

LOGGER = logging.getLogger(__name__)


class AnonCredsRevocationSetupManager(ABC):
    """Base class for automated setup of revocation."""

    @abstractmethod
    def register_events(self, event_bus: EventBus) -> None:
        """Event registration."""


class DefaultRevocationSetup(AnonCredsRevocationSetupManager):
    """Manager for automated setup of revocation support.

    This manager models a state machine for the revocation setup process where
    the transitions are triggered by the `finished` event of the previous
    artifact. The state machine is as follows:

    [*] --> Cred Def
    Cred Def --> Rev Reg Def
    Rev Reg Def --> Rev List
    Rev List --> [*]

    This implementation of an AnonCredsRevocationSetupManager will create two
    revocation registries for each credential definition supporting revocation;
    one that is active and one that is pending. When the active registry fills,
    the pending registry will be activated and a new pending registry will be
    created. This will continue indefinitely.

    This hot-swap approach to revocation registry management allows for
    issuance operations to be performed without a delay for registry
    creation.
    """

    REGISTRY_TYPE = "CL_ACCUM"

    def __init__(self) -> None:
        """Init manager."""

    async def _setup_request_correlation(
        self,
        profile: Profile,
        payload: BaseEventPayload,
        event_type: str,
    ) -> tuple[str, dict]:
        """Set up correlation ID and event storage for request handlers.

        Args:
            profile: The profile context
            payload: The event payload containing options
            event_type: The event type for storage

        Returns:
            tuple: (correlation_id, options_with_correlation)

        """
        # Check if this is a retry with existing correlation_id
        correlation_id = payload.options.get("correlation_id")
        if not correlation_id:
            # Generate new correlation_id for new requests
            correlation_id = generate_correlation_id()

            # Persist the request event only for new requests
            async with profile.session() as session:
                event_storage = EventStorageManager(session)

                # Calculate expiry timestamp based on current retry count
                retry_count = payload.options.get("retry_count", 0)
                expiry_timestamp = calculate_event_expiry_timestamp(retry_count)

                await event_storage.store_event_request(
                    event_type=event_type,
                    event_data=serialize_event_payload(payload),
                    correlation_id=correlation_id,
                    request_id=payload.options.get("request_id"),
                    options=payload.options,
                    expiry_timestamp=expiry_timestamp,
                )

        # Store correlation_id in options for response tracking
        options_with_correlation = payload.options.copy()
        options_with_correlation["correlation_id"] = correlation_id

        return correlation_id, options_with_correlation

    async def _handle_response_failure(
        self,
        profile: Profile,
        payload: BasePayloadWithFailure,
        event_type: str,
        correlation_id: str,
        failure_type: str,
        retry_callback: Callable[..., Awaitable[Any]],
    ) -> bool:
        """Handle failure response with retry logic.

        Args:
            profile: The profile context
            payload: The event payload containing failure info
            event_type: The event type for storage
            correlation_id: The correlation ID for tracking
            failure_type: Description of the failure type for logging
            retry_callback: Function to call for retry

        Returns:
            bool: True if retry was attempted, False if not retryable

        """
        failure = payload.failure
        error_info = failure.error_info

        # Log error details based on available failure attributes
        identifier: str = (
            getattr(failure, "cred_def_id", None)  # type: ignore[assignment]
            or getattr(failure, "rev_reg_def_id", None)
            or getattr(payload, "rev_reg_def_id", "unknown")
        )

        LOGGER.warning(
            "%s failed for %s, request_id: %s, correlation_id: %s, error: %s",
            failure_type.replace("_", " ").title(),
            identifier,
            payload.options.get("request_id"),
            correlation_id,
            error_info.error_msg,
        )

        # Implement exponential backoff retry logic
        if error_info.should_retry:
            retry_delay = calculate_exponential_backoff_delay(error_info.retry_count)

            LOGGER.info(
                "Retrying %s for %s, request_id: %s, correlation_id: %s. "
                "Attempt %d, delay %d seconds",
                failure_type.replace("_", " "),
                identifier,
                payload.options.get("request_id"),
                correlation_id,
                error_info.retry_count + 1,
                retry_delay,
            )

            await asyncio.sleep(retry_delay)

            # Update options with new retry count and update event for retry
            new_options = payload.options.copy()
            new_options["retry_count"] = error_info.retry_count + 1

            if correlation_id:
                async with profile.session() as session:
                    event_storage = EventStorageManager(session)
                    # Update the event for retry (sets state to REQUESTED)
                    await event_storage.update_event_for_retry(
                        event_type=event_type,
                        correlation_id=correlation_id,
                        error_msg=error_info.error_msg,
                        retry_count=error_info.retry_count + 1,
                        updated_options=new_options,
                    )

            # Execute retry callback
            await retry_callback(options=new_options)
            return True
        else:
            # Not retryable, update event as failed and notify issuer
            LOGGER.error(
                "Won't retry %s for %s, request_id: %s, correlation_id: %s",
                failure_type.replace("_", " "),
                identifier,
                payload.options.get("request_id"),
                correlation_id,
            )

            # Update event as failed and mark as completed
            if correlation_id:
                async with profile.session() as session:
                    event_storage = EventStorageManager(session)
                    await event_storage.update_event_response(
                        event_type=event_type,
                        correlation_id=correlation_id,
                        success=False,
                        response_data=serialize_event_payload(payload),
                        error_msg=error_info.error_msg,
                    )

            await self._notify_issuer_about_failure(
                profile=profile,
                failure_type=failure_type,
                identifier=identifier,
                error_msg=error_info.error_msg,
                options=payload.options,
            )
            return False

    async def _handle_response_success(
        self,
        profile: Profile,
        payload: BaseEventPayload,
        event_type: str,
        correlation_id: str,
        success_message: str,
    ) -> None:
        """Handle success response by updating event storage.

        Args:
            profile: The profile context
            payload: The event payload
            event_type: The event type for storage
            correlation_id: The correlation ID for tracking
            success_message: Log message for success

        """
        # Log success
        LOGGER.info(success_message)

        # Update event as successful and mark as completed
        if correlation_id:
            async with profile.session() as session:
                event_storage = EventStorageManager(session)
                await event_storage.update_event_response(
                    event_type=event_type,
                    correlation_id=correlation_id,
                    success=True,
                    response_data=serialize_event_payload(payload),
                )

    def _clean_options_for_new_request(self, options: dict) -> dict:
        """Clean options for new request by removing correlation_id.

        Each new request should have a unique correlation_id. When transitioning
        from one successful operation to the next new operation, we need to remove
        the correlation_id so that the next operation generates its own unique
        correlation_id.

        Args:
            options (dict): Original options dictionary

        Returns:
            dict: Cleaned options dictionary without correlation_id

        """
        cleaned_options = options.copy()
        cleaned_options.pop("correlation_id", None)
        return cleaned_options

    def register_events(self, event_bus: EventBus) -> None:
        """Register event listeners."""
        # On cred def, request creation and registration of a revocation registry
        event_bus.subscribe(CredDefFinishedEvent.event_topic, self.on_cred_def)

        # On registry create requested, create and register a revocation registry
        event_bus.subscribe(
            RevRegDefCreateRequestedEvent.event_topic, self.on_registry_create_requested
        )
        # On registry create response, emit event to store the revocation registry
        event_bus.subscribe(
            RevRegDefCreateResponseEvent.event_topic, self.on_registry_create_response
        )

        # On registry store requested, store the revocation registry
        event_bus.subscribe(
            RevRegDefStoreRequestedEvent.event_topic, self.on_registry_store_requested
        )
        # On store success, emit rev reg finished event, and requests backup registry
        event_bus.subscribe(
            RevRegDefStoreResponseEvent.event_topic, self.on_registry_store_response
        )

        # Rev list finished event will notify the issuer of successful revocations
        event_bus.subscribe(RevListFinishedEvent.event_topic, self.on_rev_list_finished)

        # On rev list create requested, create and register a revocation list
        event_bus.subscribe(
            RevListCreateRequestedEvent.event_topic, self.on_rev_list_create_requested
        )
        # On successful rev list creation, emit store rev list request event
        event_bus.subscribe(
            RevListCreateResponseEvent.event_topic, self.on_rev_list_create_response
        )

        # On rev list store requested, store the revocation list
        event_bus.subscribe(
            RevListStoreRequestedEvent.event_topic, self.on_rev_list_store_requested
        )
        # On store success, emit set active registry event, if it is the first registry
        event_bus.subscribe(
            RevListStoreResponseEvent.event_topic, self.on_rev_list_store_response
        )

        # On set active registry requested, set the active registry
        event_bus.subscribe(
            RevRegActivationRequestedEvent.event_topic,
            self.on_registry_activation_requested,
        )
        # On successful registry activation, this completes the revocation setup
        event_bus.subscribe(
            RevRegActivationResponseEvent.event_topic,
            self.on_registry_activation_response,
        )

        event_bus.subscribe(
            RevRegFullDetectedEvent.event_topic,
            self.on_registry_full_detected,
        )
        event_bus.subscribe(
            RevRegFullHandlingResponseEvent.event_topic,
            self.on_registry_full_handling_response,
        )

    async def on_cred_def(self, profile: Profile, event: CredDefFinishedEvent) -> None:
        """Handle cred def finished."""
        payload = event.payload

        if payload.support_revocation:
            revoc = AnonCredsRevocation(profile)

            # Generate a new request_id for this revocation registry workflow
            request_id = generate_request_id()
            options = self._clean_options_for_new_request(payload.options)
            options["request_id"] = request_id

            LOGGER.info(
                "Starting revocation registry workflow for cred_def_id: %s, "
                "request_id: %s",
                payload.cred_def_id,
                request_id,
            )

            # Emit event to request creation and registration of a revocation registry
            # This automates the creation of a backup registry and accompanying resources
            await revoc.emit_create_revocation_registry_definition_event(
                issuer_id=payload.issuer_id,
                cred_def_id=payload.cred_def_id,
                registry_type=self.REGISTRY_TYPE,
                max_cred_num=payload.max_cred_num,
                tag=FIRST_REGISTRY_TAG,
                options=options,
            )

            if event.payload.options.get("wait_for_revocation_setup"):
                # Wait for registry activation, if configured to do so
                await revoc.wait_for_active_revocation_registry(payload.cred_def_id)

    async def on_registry_create_requested(
        self, profile: Profile, event: RevRegDefCreateRequestedEvent
    ) -> None:
        """Handle registry creation request."""
        payload = event.payload
        revoc = AnonCredsRevocation(profile)

        correlation_id, options_with_correlation = await self._setup_request_correlation(
            profile,
            payload,  # type: ignore[arg-type]
            RECORD_TYPE_REV_REG_DEF_CREATE_EVENT,
        )

        LOGGER.debug(
            "Handling registry creation request for cred_def_id: %s, tag: %s, "
            "request_id: %s, correlation_id: %s",
            payload.cred_def_id,
            payload.tag,
            payload.options.get("request_id"),
            correlation_id,
        )

        await asyncio.shield(
            revoc.create_and_register_revocation_registry_definition(
                issuer_id=payload.issuer_id,
                cred_def_id=payload.cred_def_id,
                registry_type=payload.registry_type,
                tag=payload.tag,
                max_cred_num=payload.max_cred_num,
                options=options_with_correlation,
            )
        )

    async def on_registry_create_response(
        self, profile: Profile, event: RevRegDefCreateResponseEvent
    ) -> None:
        """Handle registry creation response."""
        payload = event.payload
        correlation_id: str = payload.options.get("correlation_id", "")

        if not correlation_id:  # pragma: no cover
            LOGGER.warning("No correlation_id found for rev reg def create response")

        if failure := payload.failure:
            # Define retry callback for registry creation
            async def retry_registry_creation(options):  # pragma: no cover
                revoc = AnonCredsRevocation(profile)
                await revoc.emit_create_revocation_registry_definition_event(
                    issuer_id=failure.issuer_id,
                    cred_def_id=failure.cred_def_id,
                    registry_type=failure.registry_type,
                    tag=failure.tag,
                    max_cred_num=failure.max_cred_num,
                    options=options,
                )

            await self._handle_response_failure(
                profile=profile,
                payload=payload,  # type: ignore[arg-type]
                event_type=RECORD_TYPE_REV_REG_DEF_CREATE_EVENT,
                correlation_id=correlation_id,
                failure_type="registry_create",
                retry_callback=retry_registry_creation,
            )
        else:
            if not payload.rev_reg_def_result or not payload.rev_reg_def:
                #  For type checks; should never happen
                LOGGER.error("Expected rev_reg_def to be present in successful response")
                return

            # Handle success
            success_message = (
                f"Registry creation succeeded for "
                f"rev_reg_def_id: {payload.rev_reg_def_result.rev_reg_def_id}, "
                f"request_id: {payload.options.get('request_id')}, "
                f"correlation_id: {correlation_id}"
            )

            await self._handle_response_success(
                profile=profile,
                payload=payload,  # type: ignore[arg-type]
                event_type=RECORD_TYPE_REV_REG_DEF_CREATE_EVENT,
                correlation_id=correlation_id,
                success_message=success_message,
            )

            # Emit next event in chain - store request event
            revoc = AnonCredsRevocation(profile)
            await revoc.emit_store_revocation_registry_definition_event(
                rev_reg_def=payload.rev_reg_def,
                rev_reg_def_result=payload.rev_reg_def_result,
                options=self._clean_options_for_new_request(payload.options),
            )

    async def on_registry_store_requested(
        self, profile: Profile, event: RevRegDefStoreRequestedEvent
    ) -> None:
        """Handle registry store request."""
        payload = event.payload
        revoc = AnonCredsRevocation(profile)

        _, options_with_correlation = await self._setup_request_correlation(
            profile,
            payload,  # type: ignore[arg-type]
            RECORD_TYPE_REV_REG_DEF_STORE_EVENT,
        )

        await revoc.handle_store_revocation_registry_definition_request(
            rev_reg_def_result=payload.rev_reg_def_result,
            options=options_with_correlation,
        )

    async def on_registry_store_response(
        self, profile: Profile, event: RevRegDefStoreResponseEvent
    ) -> None:
        """Handle registry store response."""
        payload = event.payload

        # Update the persisted event with response information
        correlation_id: str = payload.options.get("correlation_id", "")
        if not correlation_id:  # pragma: no cover
            LOGGER.warning("No correlation_id found for rev reg def store response")

        if payload.failure:
            # Define retry callback for registry store
            async def retry_registry_store(options):  # pragma: no cover
                revoc = AnonCredsRevocation(profile)
                await revoc.handle_store_revocation_registry_definition_request(
                    rev_reg_def_result=payload.rev_reg_def_result,
                    options=options,
                )

            await self._handle_response_failure(
                profile=profile,
                payload=payload,  # type: ignore[arg-type]
                event_type=RECORD_TYPE_REV_REG_DEF_STORE_EVENT,
                correlation_id=correlation_id,
                failure_type="registry_store",
                retry_callback=retry_registry_store,
            )
        else:
            # Handle success
            success_message = (
                f"Registry store succeeded for rev_reg_def_id: {payload.rev_reg_def_id}, "
                f"tag: {payload.tag}, request_id: {payload.options.get('request_id')}, "
                f"correlation_id: {correlation_id}"
            )

            await self._handle_response_success(
                profile=profile,
                payload=payload,  # type: ignore[arg-type]
                event_type=RECORD_TYPE_REV_REG_DEF_STORE_EVENT,
                correlation_id=correlation_id,
                success_message=success_message,
            )

            # Emit finished event
            revoc = AnonCredsRevocation(profile)
            state = payload.rev_reg_def_result.revocation_registry_definition_state.state
            if state == STATE_FINISHED:
                await revoc.emit_create_and_register_revocation_list_event(
                    rev_reg_def_id=payload.rev_reg_def_id,
                    options=self._clean_options_for_new_request(payload.options),
                )
            else:  # pragma: no cover
                LOGGER.warning(
                    "Revocation registry definition %s not finished; has state %s, "
                    "request_id: %s, correlation_id: %s",
                    payload.rev_reg_def_id,
                    state,
                    payload.options.get("request_id"),
                    payload.options.get("correlation_id"),
                )

            # If this is the first registry, trigger creation of backup registry
            if payload.tag == FIRST_REGISTRY_TAG:
                # Generate new request_id for backup registry workflow
                backup_request_id = generate_request_id()
                backup_options = self._clean_options_for_new_request(payload.options)
                backup_options["request_id"] = backup_request_id

                LOGGER.info(
                    "First registry stored successfully, "
                    "requesting creation of backup registry for cred_def_id: %s, "
                    "original request_id: %s, new backup request_id: %s",
                    payload.rev_reg_def.cred_def_id,
                    payload.options.get("request_id"),
                    backup_request_id,
                )

                await revoc.emit_create_revocation_registry_definition_event(
                    issuer_id=payload.rev_reg_def.issuer_id,
                    cred_def_id=payload.rev_reg_def.cred_def_id,
                    registry_type=payload.rev_reg_def.type,
                    tag=revoc._generate_backup_registry_tag(),
                    max_cred_num=payload.rev_reg_def.value.max_cred_num,
                    options=backup_options,
                )

    async def on_rev_list_create_requested(
        self, profile: Profile, event: RevListCreateRequestedEvent
    ) -> None:
        """Handle revocation list creation request."""
        payload = event.payload
        revoc = AnonCredsRevocation(profile)

        correlation_id, options_with_correlation = await self._setup_request_correlation(
            profile,
            payload,  # type: ignore[arg-type]
            RECORD_TYPE_REV_LIST_CREATE_EVENT,
        )

        LOGGER.debug(
            "Handling revocation list creation request for rev_reg_def_id: %s, "
            "request_id: %s, correlation_id: %s",
            payload.rev_reg_def_id,
            payload.options.get("request_id"),
            correlation_id,
        )

        await asyncio.shield(
            revoc.create_and_register_revocation_list(
                rev_reg_def_id=payload.rev_reg_def_id,
                options=options_with_correlation,
            )
        )

    async def on_rev_list_create_response(
        self, profile: Profile, event: RevListCreateResponseEvent
    ) -> None:
        """Handle revocation list creation response."""
        payload = event.payload

        # Update the persisted event with response information
        correlation_id: str = payload.options.get("correlation_id", "")
        if not correlation_id:  # pragma: no cover
            LOGGER.warning("No correlation_id found for rev list create response")

        if payload.failure:
            # Define retry callback for rev list creation
            async def retry_rev_list_creation(options):  # pragma: no cover
                revoc = AnonCredsRevocation(profile)
                await revoc.emit_create_and_register_revocation_list_event(
                    payload.rev_reg_def_id, options
                )

            await self._handle_response_failure(
                profile=profile,
                payload=payload,  # type: ignore[arg-type]
                event_type=RECORD_TYPE_REV_LIST_CREATE_EVENT,
                correlation_id=correlation_id,
                failure_type="rev_list_create",
                retry_callback=retry_rev_list_creation,
            )
        else:
            if not payload.rev_list_result:
                #  For type checks; should never happen
                LOGGER.error(
                    "Expected rev_list_result to exist in successful create response"
                )
                return

            # Handle success
            success_message = (
                f"Revocation list creation succeeded for "
                f"rev_reg_def_id: {payload.rev_reg_def_id}, "
                f"request_id: {payload.options.get('request_id')}, "
                f"correlation_id: {correlation_id}"
            )

            await self._handle_response_success(
                profile=profile,
                payload=payload,  # type: ignore[arg-type]
                event_type=RECORD_TYPE_REV_LIST_CREATE_EVENT,
                correlation_id=correlation_id,
                success_message=success_message,
            )

            # Emit store request event
            revoc = AnonCredsRevocation(profile)
            await revoc.emit_store_revocation_list_event(
                rev_reg_def_id=payload.rev_reg_def_id,
                result=payload.rev_list_result,
                options=self._clean_options_for_new_request(payload.options),
            )

    async def on_rev_list_finished(
        self, profile: Profile, event: RevListFinishedEvent
    ) -> None:
        """Handle rev list finished."""
        await notify_revocation_published_event(
            profile, event.payload.rev_reg_id, event.payload.revoked
        )

    async def on_rev_list_store_requested(
        self, profile: Profile, event: RevListStoreRequestedEvent
    ) -> None:
        """Handle revocation list store request."""
        payload = event.payload
        revoc = AnonCredsRevocation(profile)

        correlation_id, options_with_correlation = await self._setup_request_correlation(
            profile,
            payload,  # type: ignore[arg-type]
            RECORD_TYPE_REV_LIST_STORE_EVENT,
        )

        LOGGER.debug(
            "Handling revocation list store request for rev_reg_def_id: %s, "
            "request_id: %s, correlation_id: %s",
            payload.rev_reg_def_id,
            payload.options.get("request_id"),
            correlation_id,
        )

        await revoc.handle_store_revocation_list_request(
            rev_reg_def_id=payload.rev_reg_def_id,
            result=payload.result,
            options=options_with_correlation,
        )

    async def on_rev_list_store_response(
        self, profile: Profile, event: RevListStoreResponseEvent
    ) -> None:
        """Handle revocation list store response."""
        payload = event.payload

        # Update the persisted event with response information
        correlation_id: str = payload.options.get("correlation_id", "")
        if not correlation_id:  # pragma: no cover
            LOGGER.warning("No correlation_id found for rev list store response")

        if payload.failure:
            # Define retry callback for rev list store
            async def retry_rev_list_store(options):  # pragma: no cover
                revoc = AnonCredsRevocation(profile)
                await revoc.handle_store_revocation_list_request(
                    rev_reg_def_id=payload.rev_reg_def_id,
                    result=payload.result,
                    options=options,
                )

            await self._handle_response_failure(
                profile=profile,
                payload=payload,  # type: ignore[arg-type]
                event_type=RECORD_TYPE_REV_LIST_STORE_EVENT,
                correlation_id=correlation_id,
                failure_type="rev_list_store",
                retry_callback=retry_rev_list_store,
            )
        else:
            # Handle success
            success_message = (
                f"Revocation list store succeeded for "
                f"rev_reg_def_id: {payload.rev_reg_def_id}, "
                f"request_id: {payload.options.get('request_id')}, "
                f"correlation_id: {correlation_id}"
            )

            await self._handle_response_success(
                profile=profile,
                payload=payload,  # type: ignore[arg-type]
                event_type=RECORD_TYPE_REV_LIST_STORE_EVENT,
                correlation_id=correlation_id,
                success_message=success_message,
            )

            # If this is for the first registry, activate it
            revoc = AnonCredsRevocation(profile)
            options = self._clean_options_for_new_request(payload.options)
            first_registry = options.pop("first_registry", False)
            if first_registry:
                await revoc.emit_set_active_registry_event(
                    rev_reg_def_id=payload.rev_reg_def_id,
                    options=options,
                )

    async def on_registry_activation_requested(
        self, profile: Profile, event: RevRegActivationRequestedEvent
    ) -> None:
        """Handle registry activation request."""
        payload = event.payload
        revoc = AnonCredsRevocation(profile)

        correlation_id, options_with_correlation = await self._setup_request_correlation(
            profile,
            payload,  # type: ignore[arg-type]
            RECORD_TYPE_REV_REG_ACTIVATION_EVENT,
        )

        LOGGER.debug(
            "Handling registry activation request for rev_reg_def_id: %s, "
            "cred_def_id: %s, request_id: %s, correlation_id: %s",
            payload.rev_reg_def_id,
            payload.options.get("cred_def_id"),
            payload.options.get("request_id"),
            correlation_id,
        )

        await revoc.handle_activate_registry_request(
            rev_reg_def_id=payload.rev_reg_def_id,
            options=options_with_correlation,
        )

    async def on_registry_activation_response(
        self, profile: Profile, event: RevRegActivationResponseEvent
    ) -> None:
        """Handle registry activation response."""
        payload = event.payload

        # Update the persisted event with response information
        correlation_id: str = payload.options.get("correlation_id", "")
        if not correlation_id:  # pragma: no cover
            LOGGER.warning("No correlation_id found for rev reg def activation response")

        if payload.failure:

            async def retry_activation(options):  # pragma: no cover
                revoc = AnonCredsRevocation(profile)
                await revoc.emit_set_active_registry_event(
                    payload.rev_reg_def_id, options
                )

            await self._handle_response_failure(
                profile=profile,
                payload=payload,  # type: ignore[arg-type]
                event_type=RECORD_TYPE_REV_REG_ACTIVATION_EVENT,
                correlation_id=correlation_id,
                failure_type="registry_activation",
                retry_callback=retry_activation,
            )
        else:
            # Handle success
            LOGGER.info(
                "Registry activation succeeded for rev_reg_def_id: %s, "
                "cred_def_id: %s, request_id: %s, correlation_id: %s",
                payload.rev_reg_def_id,
                payload.options.get("cred_def_id"),
                payload.options.get("request_id"),
                payload.options.get("correlation_id"),
            )

            # Update event as successful and mark as completed
            if correlation_id:
                async with profile.session() as session:
                    event_storage = EventStorageManager(session)
                    await event_storage.update_event_response(
                        event_type=RECORD_TYPE_REV_REG_ACTIVATION_EVENT,
                        correlation_id=correlation_id,
                        success=True,
                        response_data=serialize_event_payload(payload),
                    )

            # Check if this request was part of full registry handling; then create backup
            if payload.options.get("cred_def_id") and payload.options.get(
                "old_rev_reg_def_id"
            ):
                # Get the registry definition to extract issuer details
                revoc = AnonCredsRevocation(profile)
                rev_reg_def = await revoc.get_created_revocation_registry_definition(
                    payload.rev_reg_def_id
                )

                if rev_reg_def:
                    # Create new backup registry with same request_id
                    backup_options = self._clean_options_for_new_request(payload.options)
                    backup_options["request_id"] = payload.options.get("request_id")

                    LOGGER.debug(
                        "Emitting event to create new backup registry for "
                        "cred def id %s, request_id: %s, correlation_id: %s",
                        payload.options["cred_def_id"],
                        payload.options.get("request_id"),
                        payload.options.get("correlation_id"),
                    )
                    await revoc.emit_create_revocation_registry_definition_event(
                        issuer_id=rev_reg_def.issuer_id,
                        cred_def_id=payload.options["cred_def_id"],
                        registry_type=rev_reg_def.type,
                        tag=revoc._generate_backup_registry_tag(),
                        max_cred_num=rev_reg_def.value.max_cred_num,
                        options=backup_options,
                    )
                else:
                    LOGGER.error(
                        "Could not retrieve registry definition %s for creating backup",
                        payload.rev_reg_def_id,
                    )
                    await self._notify_issuer_about_failure(
                        profile=profile,
                        failure_type="registry_activation",
                        identifier=payload.rev_reg_def_id,
                        error_msg="Could not retrieve registry definition for creating backup",  # noqa: E501
                        options=payload.options,
                    )

    async def on_registry_full_detected(
        self, profile: Profile, event: RevRegFullDetectedEvent
    ) -> None:
        """Handle registry full detection."""
        payload = event.payload
        revoc = AnonCredsRevocation(profile)

        # Check if this is a retry with existing correlation_id
        correlation_id = payload.options.get("correlation_id")
        if not correlation_id:
            # Generate new correlation_id and request_id for new full registry handling
            correlation_id = generate_correlation_id()

            # Generate new request_id for full registry handling workflow
            if "request_id" not in payload.options:
                full_handling_request_id = generate_request_id()
                payload.options["request_id"] = full_handling_request_id

                LOGGER.info(
                    "Starting full registry handling workflow for rev_reg_def_id: %s, "
                    "cred_def_id: %s, request_id: %s",
                    payload.rev_reg_def_id,
                    payload.cred_def_id,
                    full_handling_request_id,
                )

            # Persist the request event only for new requests
            async with profile.session() as session:
                event_storage = EventStorageManager(session)

                # Calculate expiry timestamp based on current retry count
                retry_count = payload.options.get("retry_count", 0)
                expiry_timestamp = calculate_event_expiry_timestamp(retry_count)

                await event_storage.store_event_request(
                    event_type=RECORD_TYPE_REV_REG_FULL_HANDLING_EVENT,
                    event_data=serialize_event_payload(payload),
                    correlation_id=correlation_id,
                    request_id=payload.options.get("request_id"),
                    options=payload.options,
                    expiry_timestamp=expiry_timestamp,
                )

        LOGGER.info(
            "Full registry detected for cred_def_id: %s. Full rev_reg_def_id: %s. "
            "request_id: %s. correlation_id: %s",
            payload.cred_def_id,
            payload.rev_reg_def_id,
            payload.options.get("request_id"),
            correlation_id,
        )

        # Store correlation_id in options for response tracking
        options_with_correlation = payload.options.copy()
        options_with_correlation["correlation_id"] = correlation_id

        # Start the full registry handling process
        await revoc.handle_full_registry_event(
            rev_reg_def_id=payload.rev_reg_def_id,
            cred_def_id=payload.cred_def_id,
            options=options_with_correlation,
        )

    async def on_registry_full_handling_response(
        self, profile: Profile, event: RevRegFullHandlingResponseEvent
    ) -> None:
        """Handle registry full handling completed."""
        payload = event.payload

        # Update the persisted event with response information
        correlation_id: str = payload.options.get("correlation_id", "")
        if not correlation_id:  # pragma: no cover
            LOGGER.warning("No correlation_id found for full registry handling response")

        if payload.failure:
            # Define retry callback for full registry handling
            async def retry_full_registry_handling(options):  # pragma: no cover
                revoc = AnonCredsRevocation(profile)
                await revoc.handle_full_registry_event(
                    rev_reg_def_id=payload.old_rev_reg_def_id,
                    cred_def_id=payload.cred_def_id,
                    options=options,
                )

            await self._handle_response_failure(
                profile=profile,
                payload=payload,  # type: ignore[arg-type]
                event_type=RECORD_TYPE_REV_REG_FULL_HANDLING_EVENT,
                correlation_id=correlation_id,
                failure_type="full_registry_handling",
                retry_callback=retry_full_registry_handling,
            )

        else:
            # Handle success
            success_message = (
                f"Full registry handling response. "
                f"Old rev reg def id: {payload.old_rev_reg_def_id}, "
                f"new active rev reg def id: {payload.new_active_rev_reg_def_id}, "
                f"cred_def_id: {payload.cred_def_id}, "
                f"request_id: {payload.options.get('request_id')}, "
                f"correlation_id: {correlation_id}"
            )

            await self._handle_response_success(
                profile=profile,
                payload=payload,  # type: ignore[arg-type]
                event_type=RECORD_TYPE_REV_REG_FULL_HANDLING_EVENT,
                correlation_id=correlation_id,
                success_message=success_message,
            )

    # Helper methods for error handling and notifications
    async def _notify_issuer_about_failure(
        self,
        profile: Profile,
        failure_type: str,
        identifier: str,
        error_msg: str,
        options: dict,
    ) -> None:
        """Notify issuer about a failure that couldn't be automatically recovered.

        Args:
            profile (Profile): The profile context
            failure_type (str): Type of failure (e.g. "registry_creation", "tails_upload")
            identifier (str): Identifier for the failed operation
            error_msg (str): Error message
            options (dict): Options context

        """
        LOGGER.error(
            f"MANUAL INTERVENTION REQUIRED: {failure_type} failed for {identifier} "
            f"after maximum retries. Error: {error_msg}. Options: {options}"
        )

        event_bus = profile.inject_or(EventBus)
        if event_bus:
            await event_bus.notify(
                profile=profile,
                event=Event(
                    topic=INTERVENTION_REQUIRED_EVENT,
                    payload=InterventionRequiredPayload(
                        point_of_failure=failure_type,
                        error_msg=error_msg,
                        identifier=identifier,
                        options=options,
                    ),
                ),
            )
        else:
            LOGGER.error(
                "Could not notify issuer %s about failure %s for %s",
                profile.name,
                failure_type,
                identifier,
            )



"""Module docstring."""

CATEGORY = "connection_request"

SCHEMAS = {
    "sqlite": [
        """
        CREATE TABLE IF NOT EXISTS connection_request_v0_1 (
            id INTEGER PRIMARY KEY AUTOINCREMENT,
            item_id INTEGER NOT NULL,
            item_name TEXT NOT NULL,
            connection_id TEXT NOT NULL,
            message_id TEXT,
            type TEXT,
            label TEXT,
            image_url TEXT,
            did TEXT,
            thread_pthid TEXT,
            did_doc TEXT,  -- JSON string of did_doc~attach
            created_at DATETIME DEFAULT CURRENT_TIMESTAMP,
            updated_at DATETIME DEFAULT CURRENT_TIMESTAMP,
            FOREIGN KEY (item_id) REFERENCES items(id) ON DELETE CASCADE ON UPDATE CASCADE
        );
        """,
        """
        CREATE INDEX IF NOT EXISTS idx_connection_request_item_id_v0_1
        ON connection_request_v0_1 (item_id);
        """,
        """
        CREATE INDEX IF NOT EXISTS idx_connection_request_message_id_v0_1
        ON connection_request_v0_1 (message_id);
        """,
        """
        CREATE INDEX IF NOT EXISTS idx_connection_request_did_v0_1
        ON connection_request_v0_1 (did);
        """,
        """
        CREATE INDEX IF NOT EXISTS idx_connection_request_thread_pthid_v0_1
        ON connection_request_v0_1 (thread_pthid);
        """,
        """
        CREATE TRIGGER IF NOT EXISTS trg_update_connection_request_timestamp_v0_1
        AFTER UPDATE ON connection_request_v0_1
        FOR EACH ROW
        BEGIN
            UPDATE connection_request_v0_1
            SET updated_at = CURRENT_TIMESTAMP
            WHERE id = OLD.id;
        END;
        """,
    ],
    "postgresql": [
        """
        CREATE TABLE IF NOT EXISTS connection_request_v0_1 (
            id SERIAL PRIMARY KEY,
            item_id INTEGER NOT NULL,
            item_name TEXT NOT NULL,
            connection_id TEXT NOT NULL,
            message_id TEXT,
            type TEXT,
            label TEXT,
            image_url TEXT,
            did TEXT,
            thread_pthid TEXT,
            did_doc TEXT,  -- JSON string of did_doc~attach
            created_at TIMESTAMP WITH TIME ZONE DEFAULT CURRENT_TIMESTAMP,
            updated_at TIMESTAMP WITH TIME ZONE DEFAULT CURRENT_TIMESTAMP,
            CONSTRAINT fk_item_id FOREIGN KEY (item_id)
            REFERENCES items(id) ON DELETE CASCADE ON UPDATE CASCADE
        );
        """,
        """
        CREATE INDEX IF NOT EXISTS idx_connection_request_item_id_v0_1
        ON connection_request_v0_1 (item_id);
        """,
        """
        CREATE INDEX IF NOT EXISTS idx_connection_request_message_id_v0_1
        ON connection_request_v0_1 (message_id);
        """,
        """
        CREATE INDEX IF NOT EXISTS idx_connection_request_did_v0_1
        ON connection_request_v0_1 (did);
        """,
        """
        CREATE INDEX IF NOT EXISTS idx_connection_request_thread_pthid_v0_1
        ON connection_request_v0_1 (thread_pthid);
        """,
        """
        CREATE OR REPLACE FUNCTION update_connection_request_timestamp_v0_1()
        RETURNS TRIGGER AS $$
        BEGIN
            IF NEW.updated_at IS NULL THEN
                NEW.updated_at = CURRENT_TIMESTAMP;
            END IF;
            RETURN NEW;
        END;
        $$ LANGUAGE plpgsql;
        """,
        """
        CREATE TRIGGER trg_update_connection_request_timestamp_v0_1
        BEFORE UPDATE ON connection_request_v0_1
        FOR EACH ROW
        EXECUTE FUNCTION update_connection_request_timestamp_v0_1();
        """,
    ],
    "mssql": [
        """
        CREATE TABLE connection_request_v0_1 (
            id INT IDENTITY(1,1) PRIMARY KEY,
            item_id INT NOT NULL,
            item_name NVARCHAR(MAX) NOT NULL,
            connection_id NVARCHAR(255) NOT NULL,
            message_id NVARCHAR(255),
            type NVARCHAR(255),
            label NVARCHAR(MAX),
            image_url NVARCHAR(MAX),
            did NVARCHAR(255),
            thread_pthid NVARCHAR(255),
            did_doc NVARCHAR(MAX),  -- JSON string of did_doc~attach
            created_at DATETIME2 DEFAULT SYSDATETIME(),
            updated_at DATETIME2 DEFAULT SYSDATETIME(),
            CONSTRAINT fk_item_id FOREIGN KEY (item_id)
            REFERENCES items(id) ON DELETE CASCADE ON UPDATE CASCADE
        );
        """,
        """
        CREATE NONCLUSTERED INDEX idx_connection_request_item_id_v0_1
        ON connection_request_v0_1 (item_id);
        """,
        """
        CREATE NONCLUSTERED INDEX idx_connection_request_message_id_v0_1
        ON connection_request_v0_1 (message_id);
        """,
        """
        CREATE NONCLUSTERED INDEX idx_connection_request_did_v0_1
        ON connection_request_v0_1 (did);
        """,
        """
        CREATE NONCLUSTERED INDEX idx_connection_request_thread_pthid_v0_1
        ON connection_request_v0_1 (thread_pthid);
        """,
        """
        CREATE TRIGGER trg_update_connection_request_timestamp_v0_1
        ON connection_request_v0_1
        AFTER UPDATE
        AS
        BEGIN
            UPDATE connection_request_v0_1
            SET updated_at = SYSDATETIME()
            FROM connection_request_v0_1
            INNER JOIN inserted ON connection_request_v0_1.id = inserted.id
            WHERE inserted.updated_at IS NULL;
        END;
        """,
    ],
}


DROP_SCHEMAS = {
    "sqlite": [
        "DROP TRIGGER IF EXISTS trg_update_connection_request_timestamp_v0_1;",
        "DROP INDEX IF EXISTS idx_connection_request_thread_pthid_v0_1;",
        "DROP INDEX IF EXISTS idx_connection_request_did_v0_1;",
        "DROP INDEX IF EXISTS idx_connection_request_message_id_v0_1;",
        "DROP INDEX IF EXISTS idx_connection_request_item_id_v0_1;",
        "DROP TABLE IF EXISTS connection_request_v0_1;",
    ],
    "postgresql": [
        """
        DROP TRIGGER IF EXISTS trg_update_connection_request_timestamp_v0_1
        ON connection_request_v0_1;
        """,
        "DROP FUNCTION IF EXISTS update_connection_request_timestamp_v0_1 CASCADE;",
        "DROP INDEX IF EXISTS idx_connection_request_thread_pthid_v0_1;",
        "DROP INDEX IF EXISTS idx_connection_request_did_v0_1;",
        "DROP INDEX IF EXISTS idx_connection_request_message_id_v0_1;",
        "DROP INDEX IF EXISTS idx_connection_request_item_id_v0_1;",
        "DROP TABLE IF EXISTS connection_request_v0_1 CASCADE;",
    ],
    "mssql": [
        "DROP TRIGGER IF EXISTS trg_update_connection_request_timestamp_v0_1;",
        """
        DROP INDEX IF EXISTS idx_connection_request_thread_pthid_v0_1
        ON connection_request_v0_1;
        """,
        """
        DROP INDEX IF EXISTS idx_connection_request_did_v0_1
        ON connection_request_v0_1;
        """,
        """
        DROP INDEX IF EXISTS idx_connection_request_message_id_v0_1
        ON connection_request_v0_1;
        """,
        """
        DROP INDEX IF EXISTS idx_connection_request_item_id_v0_1
        ON connection_request_v0_1;
        """
        "DROP TABLE IF EXISTS connection_request_v0_1;",
    ],
}


COLUMNS = [
    "message_id",
    "connection_id",
    "type",
    "label",
    "image_url",
    "did",
    "thread_pthid",
    "did_doc",
]


# Sample data structure:
# {
#     "@type": "https://didcomm.org/didexchange/1.1/request",
#     "@id": "b7958c6e-b5fd-46cb-9214-bb2490e97c9e",
#     "~thread": {"pthid": "c314ba37-b375-4022-a2d2-3e44eee7eb75"},
#     "label": "My Wallet - 0655",
#     "did": "did:peer:1zQmdGpc4Tc6gvYvEy1HtDzaXaRGetXTvMki6jm6DLSsK62L",
#     "did_doc~attach": {
#         "@id": "6864e554-658f-4b79-a6d4-9e27477d53cc",
#         "mime-type": "application/json",
#         "data": {
#             "base64": "eyJAY29udGV4dCI6WyJodHRwczovL3czaWQub3JnL2RpZC92...",
#             "jws": {
#                 "header": {"kid": "did:key:z6MkwAuKddLDirF9BCpZDKeTZXVVnpg..."},
#                 "protected": "eyJhbGciOiJFZERTQSIsImp3ayI6eyJrdHk...",
#                 "signature": "R1Cu4JlCvkJg_ToJrd3aRBfOjPFaJ9ue5Oit37hBR0c..."
#             }
#         }
#     }
# }}



"""Module docstring."""

from typing import Any, Dict, List

from ..tags import CompareOp, ConjunctionOp, TagName, TagQueryEncoder


class MongoTagEncoder(TagQueryEncoder):
    """MongoDB query encoder for tag-based queries."""

    def __init__(self, enc_name, enc_value):
        """Initialize the MongoTagEncoder with encoding functions."""
        self.enc_name = enc_name
        self.enc_value = enc_value
        self.query = {}

    def encode_name(self, name: TagName) -> str:
        """Encode a tag name using the provided encoding function."""
        return self.enc_name(name.value)

    def encode_value(self, value: str) -> Any:
        """Encode a tag value using the provided encoding function."""
        return self.enc_value(value)

    def encode_op_clause(
        self, op: CompareOp, enc_name: str, enc_value: Any, negate: bool
    ) -> Dict:
        """Encode a comparison operation clause with low branching."""

        # Direct handlers for equality and like
        def handle_eq(name: str, value: Any, not_: bool) -> Dict:
            return {name: {"$ne": value}} if not_ else {name: value}

        def handle_neq(name: str, value: Any, not_: bool) -> Dict:
            return {name: value} if not_ else {name: {"$ne": value}}

        def handle_like(name: str, value: Any, not_: bool) -> Dict:
            regex_clause = {"$regex": value}
            return {name: {"$not": regex_clause}} if not_ else {name: regex_clause}

        direct_dispatch = {
            CompareOp.Eq: handle_eq,
            CompareOp.Neq: handle_neq,
            CompareOp.Like: handle_like,
        }

        if op in direct_dispatch:
            return direct_dispatch[op](enc_name, enc_value, negate)

        # Range-like ops share the same shape
        range_op_map = {
            CompareOp.Gt: "$gt",
            CompareOp.Gte: "$gte",
            CompareOp.Lt: "$lt",
            CompareOp.Lte: "$lte",
        }
        mongo_op = range_op_map.get(op)
        if not mongo_op:
            raise ValueError(f"Unsupported operation: {op}")
        clause = {mongo_op: enc_value}
        return {enc_name: {"$not": clause}} if negate else {enc_name: clause}

    def encode_in_clause(
        self, enc_name: str, enc_values: List[Any], negate: bool
    ) -> Dict:
        """Encode an IN clause for multiple values."""
        if negate:
            return {enc_name: {"$nin": enc_values}}
        else:
            return {enc_name: {"$in": enc_values}}

    def encode_exist_clause(self, enc_name: str, negate: bool) -> Dict:
        """Encode an EXISTS clause."""
        exists_value = not negate
        return {enc_name: {"$exists": exists_value}}

    def encode_conj_clause(self, op: ConjunctionOp, clauses: List[Dict]) -> Dict:
        """Encode a conjunction (AND/OR) clause."""
        if not clauses:
            if op == ConjunctionOp.Or:
                return {"$or": []}
            return {}
        mongo_op = "$and" if op == ConjunctionOp.And else "$or"
        return {mongo_op: clauses}
